#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: HPC, FP, AMS
#
# This file is part of the yambopy project
#
import os
import numpy as np
from yambopy.units import ha2ev
from yambopy.tools.string import marquee
from yambopy.plot.bandstructure import YambopyBandStructure, YambopyBandStructureList
from yambopy.plot.plotting import add_fig_kwargs
from yamboparser import YamboFile
from yambopy.lattice import car_red, red_car, rec_lat, vol_lat
from yambopy.kpoints import get_path
from qepy.lattice import Path

class YamboQPDB():
    """
    Class to read yambo ndb.QP files

    These files describe the quasiparticle states calculated from yambo
    Includes the quasi-particle energies, the lifetimes and the Z factors
    """
    def __init__(self,qps):
        """
        Initialize the YamboQP class
        """
        self.qps          = qps
        self.kpoints_iku  = np.array(qps['Kpoint'])
        self.kpoint_index = np.array(qps['Kpoint_index'],dtype=int)
        self.band_index   = np.array(qps['Band'],dtype=int)
        self.e0           = np.array(qps['Eo']).real*ha2ev
        self.e            = np.array(qps['E']).real*ha2ev
        self.linewidths   = np.array(qps['E']).imag*ha2ev
        self.qpz          = np.array(qps['Z']).real
        self.spin         = False
        if 'Spin_pol' in qps.keys():
            self.spin_pol = qps['Spin_pol']
            self.spin     = True

    @property
    def eigenvalues_qp(self):
        if not hasattr(self,'_eigenvalues_qp'):
            self._eigenvalues_dft, self._eigenvalues_qp, self._lifetimes, self._z = self.get_qps()
        return self._eigenvalues_qp

    @property
    def eigenvalues_dft(self):
        if not hasattr(self,'_eigenvalues_dft'):
            self._eigenvalues_dft, self._eigenvalues_qp, self._lifetimes, self._z = self.get_qps()
        return self._eigenvalues_dft

    @property
    def lifetimes(self):
        if not hasattr(self,'_lifetimes'):
            self._eigenvalues_dft, self._eigenvalues_qp, self._lifetimes, self._z = self.get_qps()
        return self._lifetimes

    @property
    def z(self):
        if not hasattr(self,'_z'):
            self._eigenvalues_dft, self._eigenvalues_qp, self._lifetimes, self._z = self.get_qps()
        return self._z

    @classmethod
    def from_db(cls,filename='ndb.QP',folder='.'):
        """
        Create instance of this class from a ndb.QP file
        """
        cls.qp_path = folder
        db_path = os.path.join(folder,filename)
        if os.path.isfile(db_path):
            yfile = YamboFile(filename,folder)
        else:
            raise IOError('File %s not found'%db_path)
        return cls(yfile.data)
    from_db_file = from_db
    
    def get_qps(self):
        """
        Get quasiparticle energies in a list
        """
        #start arrays

        # If spin polarisation is present, arrays have one additional dimension

        # shift
        ncalculatedkpoints = self.max_kpoint - self.min_kpoint + 1
        if self.spin is True:
           eigenvalues_dft = np.zeros([ncalculatedkpoints,self.nbands,2])
           eigenvalues_qp  = np.zeros([ncalculatedkpoints,self.nbands,2])
           linewidths      = np.zeros([ncalculatedkpoints,self.nbands,2])
           z               = np.zeros([self.nkpoints,self.nbands,2])

           for ei,e0i,li,zi,ki,ni,s_pol in zip(self.e,self.e0,self.linewidths,self.qpz,self.kpoint_index,self.band_index,self.spin_pol):
                # position in array
                nkpoint = ki - self.min_kpoint 
                nband   = ni - self.min_band
                spol    = int(s_pol - 1)

                eigenvalues_dft[nkpoint,nband,spol] = e0i
                eigenvalues_qp[nkpoint,nband,spol]  = ei
                linewidths[nkpoint,nband,spol]      = li
                z[nkpoint,nband,spol]               = zi

        else:
            eigenvalues_dft = np.zeros([ncalculatedkpoints,self.nbands])
            eigenvalues_qp  = np.zeros([ncalculatedkpoints,self.nbands])
            linewidths      = np.zeros([ncalculatedkpoints,self.nbands])
            z               = np.zeros([self.nkpoints,self.nbands])

            for ei,e0i,li,zi,ki,ni in zip(self.e,self.e0,self.linewidths,self.qpz,self.kpoint_index,self.band_index):

                # position in array
                nkpoint = ki - self.min_kpoint 
                nband   = ni - self.min_band

                eigenvalues_dft[nkpoint,nband] = e0i
                eigenvalues_qp[nkpoint,nband]  = ei
                linewidths[nkpoint,nband]      = li
                z[nkpoint,nband]               = zi
        return eigenvalues_dft, eigenvalues_qp, linewidths, z


    def get_filtered_qps(self,min_band=None,max_band=None):
        """Return selected QP energies as a flat list"""
        e0=[]; qp=[]; lw=[]
        for ei,e0i,li,ki,ni in zip(self.e,self.e0,self.linewidths,self.kpoint_index,self.band_index):
            if min_band and ni < min_band: continue
            if max_band and ni > max_band: continue
            e0.append(e0i)
            qp.append(ei)
            lw.append(lw)
        return e0,qp,lw

    def get_direct_gaps(self,valence):
        """
        Compute the QP and DFT gaps
        
        Arguments:
            valence: number of bands in the valence
        """
        na = np.newaxis
        shifted_valence = valence-self.min_band+1
 
        #direct gap
        dft_jdos = self.eigenvalues_dft[:,na,shifted_valence:]-self.eigenvalues_dft[:,:shifted_valence,na]
        qp_jdos  = self.eigenvalues_qp[:,na,shifted_valence:] -self.eigenvalues_qp[:,:shifted_valence,na]
        direct_dft_gap = np.min(dft_jdos)
        direct_qp_gap  = np.min(qp_jdos)

        #indirect gap
        #TODO take the min and max of the VBM and CBM
        return direct_dft_gap, direct_qp_gap

    def get_scissor(self,valence,verbose=1):
        """
        Compute the scissor operator replicating the QP corrections
        
        Arguments:
            valence: number of bands in the valence
        """
        from scipy import stats
        lines = []; app = lines.append
        #valence
        e0, eqp, lw = self.get_filtered_qps(self.min_band,valence)
        vslope, vintercept, r_value, p_value, std_err = stats.linregress(e0,eqp)
        app('valence bands:')
        app('slope:     {}'.format(vslope))
        app('intercept: {}'.format(vintercept))
        app('r_value:   {}'.format(r_value))
        app('p_value:   {}'.format(p_value))
        app('std_err:   {}'.format(std_err))
       
        #conduction
        e0, eqp, lw = self.get_filtered_qps(valence+1,self.max_band)
        cslope, cintercept, r_value, p_value, std_err = stats.linregress(e0,eqp)
        app('\nconduction bands:')
        app('slope:     {}'.format(cslope))
        app('intercept: {}'.format(cintercept))
        app('r_value:   {}'.format(r_value))
        app('p_value:   {}'.format(p_value))
        app('std_err:   {}'.format(std_err))

        #get gaps
        direct_dft_gap,direct_qp_gap = self.get_direct_gaps(valence) 
        shift = direct_qp_gap-direct_dft_gap
        app('direct dft gap: {}'.format(direct_dft_gap))
        app('direct qp gap:  {}'.format(direct_qp_gap))

        scissor_list = [shift,cslope,vslope]
        app('\vscissor list (shift,c,v) [eV,adim,adim]: {}'.format(scissor_list))
        if verbose: print("\n".join(lines))
        return shift,cslope,vslope,cintercept,vintercept

    def plot_scissor_ax(self,ax,valence,verbose=1):
        # Create option to plot correction vs ks eigenvalues
        """
        Plot the scissor on a matplotlib axis
        """
        shift,cslope,vslope,cintercept,vintercept=self.get_scissor(valence,verbose=verbose)

        #plot qps
        ve0,vqp,_=self.get_filtered_qps(self.min_band,valence)
        ve0 = np.array(ve0)
        vqp = np.array(vqp)
        #ax.scatter(ve0,vqp)
        ce0,cqp,_=self.get_filtered_qps(valence+1,self.max_band)
        ce0 = np.array(ce0)
        cqp = np.array(cqp)
        #ax.scatter(ce0,cqp)

        #plot the fits
        vx = np.linspace(np.min(ve0),np.max(ve0),2)
        cx = np.linspace(np.min(ce0),np.max(ce0),2)
        vy = vslope*vx+vintercept
        cy = cslope*cx+cintercept
        #ax.plot(vx,vy)
        #ax.plot(cx,cy)

        return ax.scatter(ve0,vqp), ax.scatter(ce0,cqp), ax.plot(vx,vy), ax.plot(cx,cy)

    @add_fig_kwargs
    def plot_scissor(self,valence,verbose=1):
        """
        Plot the the QP energies and the scissor fit
        """
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(1,1,1)
        self.plot_scissor_ax(ax,valence,verbose=verbose)
        return fig

    def get_bs_path(self,lat,path,debug=False,**kwargs):
        """Get a band-structure on a path  

            Here the path is a Path object given in rlu, since it's simpler
            for the user.
        """
        
        bands_kpoints, bands_indexes, path_car = get_path(lat.car_kpoints,lat.rlat,lat.sym_car,path,debug=debug)

        # set fermi energy
        # NOT EVIDENT IN SPIN-POLARIZED SYSTEM
        #print(lat.electrons)
        #self.top_valence_band = int(lat.electrons/2)
        #print(self.top_valence_band)
        #print(self.top_valence_band-7)
        #print(self.eigenvalues_dft[:,0:(self.top_valence_band-7+2),0])
        #print(self.eigenvalues_dft[:,0:(self.top_valence_band-7+2),1])
        #exit()

        #red_bands_kpoints = car_red(bands_kpoints,lat.rlat)
        if self.spin == True:
           print('Spin polarized bands')
           ks_bandstructure_up = YambopyBandStructure(self.eigenvalues_dft[bands_indexes,:,0],bands_kpoints,kpath=path_car,**kwargs)
           ks_bandstructure_dw = YambopyBandStructure(self.eigenvalues_dft[bands_indexes,:,1],bands_kpoints,kpath=path_car,**kwargs)
           qp_bandstructure_up = YambopyBandStructure(self.eigenvalues_qp[bands_indexes,:,0], bands_kpoints,kpath=path_car,**kwargs)
           qp_bandstructure_dw = YambopyBandStructure(self.eigenvalues_qp[bands_indexes,:,1], bands_kpoints,kpath=path_car,**kwargs)

           return ks_bandstructure_up, ks_bandstructure_dw, qp_bandstructure_up, qp_bandstructure_dw
        else:
           print('no polarized bands')
           ks_bandstructure = YambopyBandStructure(self.eigenvalues_dft[bands_indexes],bands_kpoints,kpath=path_car,**kwargs)
           qp_bandstructure = YambopyBandStructure(self.eigenvalues_qp[bands_indexes] ,bands_kpoints,kpath=path_car,**kwargs)

           return ks_bandstructure, qp_bandstructure
        
    def get_bs(self,**kwargs):
        """
        Get YambopyBandStructure object with the KS and GW bands
        """
        #create bandstructure objects
        #TODO: should not be kpoints_iku but kpoints_car here
        ks_bandstructure = YambopyBandStructure(self.eigenvalues_dft,self.kpoints_iku,**kwargs)
        qp_bandstructure = YambopyBandStructure(self.eigenvalues_qp, self.kpoints_iku,**kwargs)

        return ks_bandstructure, qp_bandstructure

    def interpolate(self,lattice,path,what='QP+KS',lpratio=5,valence=None,verbose=1,**kwargs):
        """
        Interpolate the QP bandstrcture on a k-point path, requires the lattice structure with Expand=False
        """
        from yambopy.tools.skw import SkwInterpolator

        cell = (lattice.lat, lattice.red_atomic_positions, lattice.atomic_numbers)
        nelect = 0
        fermie = kwargs.pop('fermie',0)

        #consistency check with lattice points
        if len(lattice.iku_kpoints)!=len(self.kpoints_iku):
            print(len(lattice.iku_kpoints),len(self.kpoints_iku))
            raise ValueError("The QP database is not consistent with the lattice")

        if not np.isclose(lattice.iku_kpoints,self.kpoints_iku).all():
            print(lattice.iku_kpoints)
            print(self.kpoints_iku)
            raise ValueError("The QP database is not consistent with the lattice")

        #interpolate the dft eigenvalues
        kpoints = lattice.red_kpoints
        sym_rec  = lattice.sym_rec
        if not lattice.mag_syms:
            symrel = [sym for sym,trev in zip(lattice.sym_rec_red,lattice.time_rev_list) if trev==False ]
            trev_for_interp = lattice.time_rev
        # Handle special case of mag_sys + trev (e.g. SOC + ferromagnet, etc)
        elif lattice.time_rev:
            symrel = lattice.sym_rec_red
            trev_for_interp=False
        
        band_kpoints_rlu = path.get_klist()[:,:3]

        # Obtain quantities in cc needed for plot (since interpolation wants rlu)
        _, _, path_car = get_path(lattice.car_kpoints,lattice.rlat,lattice.sym_car,path)
        band_kpoints = red_car(band_kpoints_rlu,lattice.rlat)

        #interpolate KS
        ks_ebands, qp_ebands = None, None
        if 'KS' in what:
           if self.spin == True:
              print('Spin-polarized bands DFT')
              eigens_up = self.eigenvalues_dft[np.newaxis,:,:,0]
              eigens_dw = self.eigenvalues_dft[np.newaxis,:,:,1]
              skw_up = SkwInterpolator(lpratio,kpoints,eigens_up,fermie,nelect,cell,symrel,trev_for_interp,verbose=verbose)
              skw_dw = SkwInterpolator(lpratio,kpoints,eigens_dw,fermie,nelect,cell,symrel,trev_for_interp,verbose=verbose)
              dft_eigens_up_kpath = skw_up.interp_kpts(band_kpoints_rlu).eigens[0]
              dft_eigens_dw_kpath = skw_dw.interp_kpts(band_kpoints_rlu).eigens[0]

              #if valence: kwargs['fermie'] = np.max(dft_eigens_kpath[:,:valence])
              # tricky 
              ks_ebands_up = YambopyBandStructure(dft_eigens_up_kpath,band_kpoints,kpath=path_car,**kwargs)
              ks_ebands_dw = YambopyBandStructure(dft_eigens_dw_kpath,band_kpoints,kpath=path_car,**kwargs)

           else:
              print('No spin-polarized bands DFT')
              eigens  = self.eigenvalues_dft[np.newaxis,:]
              skw = SkwInterpolator(lpratio,kpoints,eigens,fermie,nelect,cell,symrel,trev_for_interp,verbose=verbose)
              #kpoints_path = path.get_klist()[:,:3]
              dft_eigens_kpath = skw.interp_kpts(band_kpoints_rlu).eigens[0]
              if valence: kwargs['fermie'] = np.max(dft_eigens_kpath[:,:valence])
              ks_ebands = YambopyBandStructure(dft_eigens_kpath,band_kpoints,kpath=path_car,**kwargs)

        #interpolate QP
        if 'QP' in what:
            if self.spin == True:
               print('Spin-polarized bands QP')
               eigens_up = self.eigenvalues_qp[np.newaxis,:,:,0]
               eigens_dw = self.eigenvalues_qp[np.newaxis,:,:,1]
               print('GW eigenvalues are sorted in ascending energy')
               #sorting
               aux_up, aux_dw = eigens_up, eigens_dw
               for ik in range(self.nkpoints):
                   eigens_up[0,ik,:], eigens_dw[0,ik,:] = sorted(aux_up[0,ik,:]), sorted(aux_dw[0,ik,:])
               #end sorting

               skw_up = SkwInterpolator(lpratio,kpoints,eigens_up,fermie,nelect,cell,symrel,trev_for_interp,verbose=verbose)
               skw_dw = SkwInterpolator(lpratio,kpoints,eigens_dw,fermie,nelect,cell,symrel,trev_for_interp,verbose=verbose)
               #kpoints_path = path.get_klist()[:,:3]
               qp_eigens_up_kpath = skw_up.interp_kpts(band_kpoints_rlu).eigens[0]
               qp_eigens_dw_kpath = skw_dw.interp_kpts(band_kpoints_rlu).eigens[0]
               #if valence: kwargs['fermie'] = np.max(dft_eigens_kpath[:,:valence])
               # tricky 
               qp_ebands_up = YambopyBandStructure(qp_eigens_up_kpath,band_kpoints,kpath=path_car,**kwargs)
               qp_ebands_dw = YambopyBandStructure(qp_eigens_dw_kpath,band_kpoints,kpath=path_car,**kwargs)

            else:
               print('No spin-polarized bands DFT')
               eigens  = self.eigenvalues_qp[np.newaxis,:]
               #sorting
               aux = eigens
               for ik in range(self.nkpoints):
                   eigens[0,ik,:] = sorted(aux[0,ik,:])
               #end sorting
               skw = SkwInterpolator(lpratio,kpoints,eigens,fermie,nelect,cell,symrel,trev_for_interp,verbose=verbose)
               #kpoints_path = path.get_klist()[:,:3]
               qp_eigens_kpath = skw.interp_kpts(band_kpoints_rlu).eigens[0]
               if valence: kwargs['fermie'] = np.max(qp_eigens_kpath[:,:valence])

               qp_ebands = YambopyBandStructure(qp_eigens_kpath,band_kpoints,kpath=path_car,**kwargs)
               #qp_ebands = YambopyBandStructure(qp_eigens_kpath,kpoints_path,kpath=path,weights=qp_z_kpath,size=0.1,**kwargs)

            qp_z_kpath = None
            if 'Z' in what:
                eigens = self.z[np.newaxis,:]
                skw = SkwInterpolator(lpratio,kpoints,eigens,fermie,nelect,cell,symrel,trev_for_interp,verbose=verbose)
                #kpoints_path = path.get_klist()[:,:3]
                qp_z_kpath = skw.interp_kpts(band_kpoints_rlu).eigens[0]
                

        if self.spin == True:
           return ks_ebands_up, ks_ebands_dw , qp_ebands_up, qp_ebands_dw
        else: 
           return ks_ebands, qp_ebands

    def interpolate_QP_corrections(self,yel_coarse,yel_dense,lpratio=20,verbose=1,**kwargs):
        """
        Interpolate QP corrections on a dense KS bandstructure.

        Input:
        * yel_coarse:   YamboElectronsDB corresponding to the YamboQPDB kgrid
        * yel_dense:    YamboElectronsDB of the dense bandstructrure

        Output
        * dense_ks_bands:   KS band structure, YambopyBandStructure object
        * dense_qp_bands:   QP band structure, YambopyBandStructure object

        """

        from yambopy.tools.skw import SkwInterpolator

        #consistency check with electrons k-points
        if len(yel_coarse.red_kpoints)!=self.nkpoints:
            print(len(yel_coarse.red_kpoints),self.nkpoints)
            raise ValueError("The QP database is not consistent with the coarse electron database")

        #consistency check with electronic bands
        if yel_dense.nbands<self.nbands:
            print(yel_dense.nbands,self.nbands)
            raise ValueError("The QP database has more bands than the electron dense bandstructure")

        nelect = 0
        fermie = kwargs.pop('fermie',0)

        kpts_coarse = yel_coarse.red_kpoints
        eigv_coarse = np.array(yel_coarse.eigenvalues_ibz)
        cell = (yel_coarse.lat, yel_coarse.red_atomic_positions, yel_coarse.atomic_numbers)
        sym_rec  = yel_coarse.sym_rec
        symrel = [sym for sym,trev in zip(yel_coarse.sym_rec_red,yel_coarse.time_rev_list) if trev==False ]
        time_rev = True

        qp_coarse = self.eigenvalues_qp
        Nkpts_coarse,Nbands_coarse = qp_coarse.shape
        ks_coarse = eigv_coarse[0][:, :Nbands_coarse]

        corrections_coarse = qp_coarse-ks_coarse
        corrections_coarse = corrections_coarse.reshape(1,Nkpts_coarse,Nbands_coarse)
        skw = SkwInterpolator(lpratio,kpts_coarse,corrections_coarse,fermie,nelect,cell,symrel,time_rev,verbose=verbose)

        kpts_dense = yel_dense.red_kpoints
        ks_dense = np.squeeze(np.array(yel_dense.eigenvalues_ibz))[:,:Nbands_coarse]
        corrections_dense = skw.interp_kpts(kpts_dense).eigens[0]
        qp_dense = corrections_dense + ks_dense

        dense_qp_bands = YambopyBandStructure(qp_dense,kpts_dense)
        dense_ks_bands = YambopyBandStructure(ks_dense,kpts_dense)

        return dense_ks_bands,dense_qp_bands


    def expand_eigenvalues(self,lattice,data=None):
        """ Expand QP values in full BZ
            - Input: YamboLatticeDB object with expanded kpts
            - Input: data as eigenvalues to expand with correct dimensions. If None, expand QP eigenvalues
        """
        nkbz   = lattice.nkpoints
        bz2ibz = lattice.kpoints_indexes

        if data is None: data = self.eigenvalues_qp

        if self.spin: eigenvalues_qp_expanded = np.zeros((nkbz,self.nbands,2))
        else:         eigenvalues_qp_expanded = np.zeros((nkbz,self.nbands))

        for ikbz in range(nkbz): eigenvalues_qp_expanded[ikbz,:] = data[bz2ibz[ikbz],:] 
        return eigenvalues_qp_expanded


    @add_fig_kwargs
    def plot_bs(self,**kwargs):
        """
        Get and plot QP bandstructure
        """
        ks_bs,qp_bs = self.get_bs(**kwargs)
        ybs = YambopyBandStructureList([ks_bs,qp_bs])
        return ybs.plot(show=False)

    @property
    def nqps(self):
        return len(self.e)

    @property
    def min_kpoint(self):
        return min(self.kpoint_index)

    @property
    def max_kpoint(self):
        return max(self.kpoint_index)

    @property
    def nbands(self):
        return self.max_band-self.min_band+1

    @property
    def min_band(self):
        return min(self.band_index)

    @property
    def max_band(self):
        return max(self.band_index)

    @property
    def nkpoints(self):
        return len(self.kpoints_iku)
    
    def __str__(self):
        lines = []; app = lines.append
        app(marquee(self.__class__.__name__))
        app( "nqps:     %d"%self.nqps)
        app( "nkpoints: %d"%self.nkpoints)
        app( "nbands:   %d"%self.nbands)
        app( "min_band: %d"%self.min_band)
        app( "max_band: %d"%self.max_band)
        if self.spin: app("Spin-polarized system.")
        return "\n".join(lines)




# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
"""
submodule to handle databases from yambo
Here we will add classes to read and manipulate the netcdf databases that yambo creates
This will allow the user to create his own scripts to manipulate the data
"""



#
# This example show how to calculate Density of States using YamboPy
#
from qepy import *
from yambopy import *
import matplotlib.pyplot as plt

save_folder='./SAVE' 
yel = YamboElectronsDB.from_db_file(folder=save_folder)
energies,dos=yel.getDOS(emin=-14.0,emax=20.0)

fig = plt.figure(figsize=(6,4))
plt.xlabel('$Energy[eV]$')
plt.ylabel('DOS$')
plt.plot(energies,dos)
plt.show()



import yaml
import argparse
import numpy as np
from math import sqrt
import xml.etree.ElementTree as ET
from qepy import xml
from qepy import bravais
from copy import deepcopy
from yambocommandline.commands import band_plots

"""
Script to produce band structure data and visualization from QE.

It reads data-file-schema.xml found in the save folders of newer pw versions.

Arguments are:
    -bi, --band_input  -> Path to yaml input file

- INPUT 
This is a yaml file in the following format, to be copy-pasted and edited:

 :: yaml

    ---
     save_dir: "PREFIX.save"
     KPTs: [[0.,0.,0.],[0.,0.5,0.],[0.5,0.5,0.],[0.5,0.,0.]]
     KPTs_labels: [G, Y, S, X]
     shift_Delta_c_v: [0.,1.,1.]

 :: 

The input parameters are:
 - save_dir: path to folder containing 'data-file-schema.xml' (in principle the QE save folder)
 - KPTs: band circuit in reduced coordinates
 - KPTs_labels: labels for the band circuit points
 - shift_Delta_c_v: k-dependent scissor shift as a list of three values (gap shift, cond. stretch, val. stretch)
"""
def read_input(inp_file):
    """
    Read input file in yaml format (see above docstring)
    """
    
    # Get input data from yaml file in dictionary form
    stream = open(inp_file, 'r')
    dictionary = yaml.safe_load(stream)
    stream.close()
    
    # Transform input data in shape used by the code
    if 'save_dir' not in dictionary: 
        raise ValueError('QE save path save_dir not found in the input file')
    else: 
        save_dir = dictionary['save_dir']
        prefix   = save_dir.split('/')[-1][:-5]
    
    if 'KPTs' not in dictionary: raise ValueError('Band circuit KPTs not found in the input file.')
    else: KPTs = dictionary['KPTs']
    
    if 'KPTs_labels' not in dictionary: raise ValueError('Kpoint labels KPTs_labels not found in the input file.')
    else: KPTs_labels = dictionary['KPTs_labels']
    
    if 'shift_Delta_c_v' in dictionary: 
        shift_Delta_c_v = np.array( dictionary['shift_Delta_c_v'] ) 
        if ( shift_Delta_c_v == np.array([0.,1.,1.]) ).all(): shift_Delta_c_v = None
    else: 
        shift_Delta_c_v = None
    
    return save_dir, prefix, KPTs, KPTs_labels, shift_Delta_c_v
    
def get_data_from_xml(input_params):
    """
    Read the data-file-schema.xml to obtain all the needed variables
    
    TO BE WRAPPED WITH A DEDICATED PWXML CLASS
    """
    save_dir, prefix, KPTs, KPTs_labels, shift_Delta_c_v = input_params
    
    # Input files from QE
    datafile  = 'data-file-schema.xml'
    datafile_xml = ET.parse( "%s/%s"%(save_dir, datafile))
    
    # Get lattice vectors
    lat_vecs = []
    lat_vecs_tags = ['a1','a2','a3']
    for tag in lat_vecs_tags:
        lat_vecs.append( xml.get_xml_data(datafile_xml,tag,as_type=float,repeated=False) )
    lat_vecs = np.array(lat_vecs)
    
    # Get kpoint number
    nkpoints = xml.get_xml_data(datafile_xml,'nk',as_type=int,repeated=True)
    
    # Get band number
    nbands   = xml.get_xml_data(datafile_xml,'nbnd',as_type=int,repeated=True)
    
    # Get kpoints cartesian coordinates
    kpts_cart, eigen, occs,= xml.get_xml_nk_bands(datafile_xml)
    
    # Check if the system is magnetic
    try:
        spin     = eval( xml.get_xml_data(datafile_xml,'lsda',as_type=str).capitalize() )
        noncolin = eval( xml.get_xml_data(datafile_xml,'noncolin',as_type=str).capitalize() )
        if spin and not noncolin: l_magnetic = True
        else: l_magnetic = False
    except ValueError: l_magnetic = False

    # Get Fermi energy
    try: fermi_e = xml.get_xml_data(datafile_xml,'fermi_energy',as_type=float)
    except ValueError: fermi_e = xml.get_xml_data(datafile_xml,'two_fermi_energies',as_type=float)[0] # majority spin Fermi energy

    # Scissor operator check
    if shift_Delta_c_v is not None:
        occ_kind = xml.get_xml_data(datafile_xml,'occupations_kind',as_type=str)
        if occ_kind!='fixed': 
            print('[WARNING]: if your system is a metal, the scissor shift will NOT work!')
    
    # Get topmost valence band (only valid for gapped systems)    
    if not l_magnetic: n_VBM = int( np.sum(occs[0]) )
    else:              n_VBM = int( np.sum(occs[0,:nbands]) ) # majority spins top valence  

    # Get eigenvalues
    eigen, s_eigen = process_bands(eigen,kpts_cart,fermi_e,shift_Delta_c_v,n_VBM)

    return lat_vecs, nkpoints, nbands, kpts_cart, eigen, s_eigen, n_VBM, l_magnetic
    
def setup_BZ_points(input_params,output_data):
    """
    Perform preliminary operations such as transform in cartesian
    coordinates and find kpath lengths.
    
    ***ASSUMING SAME NUMBER OF POINTS FOR EACH SEGMENT***)
    """
    save_dir, prefix, KPTs, KPTs_labels, shift_Delta_c_v  = input_params
    lat_vecs, nkpoints, nbands, kpts_cart, eigen, s_eigen, n_VBM, l_magnetic = output_data

    # Describe BZ symmetry lines
    nKPTs = len(KPTs)
    nkpt_per_direction = (nkpoints-1)/(nKPTs-1)

    # Metric tensor in reciprocal space
    G_r = bravais.crys_to_car(lat_vecs,reciprocal_space=True)

    # Length of path lines
    KPTs = np.array(KPTs)
    KPTdist = np.array([ KPTs[i+1]-KPTs[i] for i in range(nKPTs-1) ])

    # Transform edge kpoints from crystal to cartesian coordinates:
    #               ||x||^2 = x^T G_r x
    KPT_lengths = np.array([ sqrt(np.dot(vec, np.dot(G_r, vec))) for vec in KPTdist ])
    KPT_lengths = KPT_lengths/np.max(KPT_lengths)
    
    # Generating scaled steps
    kstps = np.zeros(nkpoints)
    points = []
    for i in range(1,nKPTs):
        n_l, n_r = np.array([i-1, i])*nkpt_per_direction
        n_l = int(n_l)
        n_r = int(n_r)
        for ik in range(n_l+1,n_r+1): kstps[ik] = kstps[ik-1]+KPT_lengths[i-1]
        points.append(kstps[n_r])

    # Get energy gaps
    gaps = find_gaps(s_eigen,n_VBM)

    # Print info
    setup_info(nKPTs,nkpt_per_direction,KPTs,KPT_lengths,nkpoints,nbands,points,gaps,shift_Delta_c_v,l_magnetic)
    
    if shift_Delta_c_v is not None: return prefix,nkpoints,nbands,kstps,KPTs_labels,points,gaps,eigen,s_eigen    
    else: return prefix,nkpoints,nbands,kstps,KPTs_labels,points,gaps,eigen
        
    
def setup_info(nKPTs,nkpt_per_direction,KPTs,KPT_lengths,nkpoints,nbands,points,gaps,scissor,l_magnetic):
    """
    Print information about the system
    """
    # BZ info
    print("=== PATH ===")
    print("directions: %d"%(nKPTs-1))
    if np.array_equal(KPTs[-1],KPTs[0]): print("(Closed path)")
    else: print("(Open path)")  
    print("ratios: ")
    print(KPT_lengths) 
    
    # Bands info 
    print("=== BANDS ===")
    print("nkpoints: %d"%nkpoints)
    print("kpoint density per direction: %d"%nkpt_per_direction)
    print("nbands: %d"%nbands)  
    if scissor is None: print("scissor shift: No")
    else: print("scissor shift: Yes")
    if gaps[0] > 1e-6 or gaps[1] >1.e-6:
        print("direct band gap:   %f eV"%gaps[0])
        print("indirect band gap: %f eV"%gaps[1])
    else:
        print("direct band gap:   %f eV"%gaps[0])
        print("indirect band gap: %f eV"%gaps[1])
        print("This is a metal.")
    
    if l_magnetic: print("The system is collinear magnetic.")

    # Symmetry points info
    print("=== PLOT ===")
    print("Internal high-symmetry points at: ")
    print(points)  

def find_gaps(eigen,n_val):
    """
    Find smallest direct and indirect band gaps
    """
    Nk,Nb = eigen.shape
    top_v, bottom_c = eigen[:,n_val-1], eigen[:,n_val]    
    dir_gaps = bottom_c - top_v
    E_dir_gap = np.min(dir_gaps)
    indir_gaps = []
    for ik in range(Nk):
        for ip in range(Nk):
            if ip!=ik: indir_gaps.append(bottom_c[ip]-top_v[ik])
    indir_gaps = np.array(indir_gaps)
    E_ind_gap = np.min(indir_gaps)
    return E_dir_gap, E_ind_gap
            
def process_bands(eigen,kpts_cart,fermi_e,shift_Delta_c_v,n_val):
    """
    Fix Fermi level to zero, apply scissor shift, return useful dictionary for plotting 
    
    Output:
        eigenvalues and shifted eigenvalues (equal if scissor not applied)
    """
    # Rescale in eV and shift Fermi level to zero
    ha2ev  = 27.211396132
    processed_eigen  = ha2ev*(eigen-fermi_e)              

    # Apply scissor shift
    shifted_eigen = deepcopy(processed_eigen)
    if shift_Delta_c_v is not None:
        scissor = shift_Delta_c_v
        top_v, bottom_c = processed_eigen[:,n_val-1], processed_eigen[:,n_val]
        #ind_k_dir_gap = np.argmin(bottom_c-top_v)
        #ev_max, ec_min = top_v[ind_k_dir_gap], bottom_c[ind_k_dir_gap]
        ev_max, ec_min = np.max(top_v), np.min(bottom_c) # The scissor is applied to the indirect gap
        for ik in range( len(kpts_cart) ):
            for ib in range( len(processed_eigen[0]) ):
                if ib<n_val: shifted_eigen[ik][ib] = ev_max-(ev_max-processed_eigen[ik][ib])*scissor[2]
                else:        shifted_eigen[ik][ib] = ec_min+scissor[0]+(processed_eigen[ik][ib]-ec_min)*scissor[1]

    return processed_eigen, shifted_eigen

def launch_plot(data,plt_type='bands',out_name=None,erange=None,show=True):
    """
    Plot driver. 
    
    Calls band_plots with the appropriate plt_type
    """
    
    band_plots.plot_driver(data,plt_type,out_name=out_name,erange=erange,show=show)

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Generate band plot')
    parser.add_argument('-bi','--band_input', type=str,help='<Required> Path to input file', required=True)
    args = parser.parse_args()

    inp = args.band_input 
    
    # Read user input file
    input_params = read_input(inp)
    
    # Read data from .xml file
    output_data = get_data_from_xml(input_params)
    
    # Plot setup
    data_to_plot = setup_BZ_points(input_params,output_data)
    
    # Produce the plot
    launch_plot(data_to_plot,plt_type='bands',out_name=None,erange=4.,show=True)
    
    





#from __future__ import print_function, division
from qepy import *
from yambopy import *
import matplotlib.pyplot as plt
import os

ymin,ymax = -2,8
npoints = 20
bse_dir = 'bse-exc'
#bse_dir = 'bse-magn'
fig = plt.figure(figsize=(4,6))
ax  = fig.add_axes( [ 0.15, 0.15, 0.80, 0.80 ])

# Define path in reduced coordinates using Class Path

path = Path([ [[  0.0,  0.0,  0.0],r'$\Gamma$'],
              [[  0.5,  0.0,  0.0],'M'],
              [[1./3.,1./3.,  0.0],'K'],
              [[  0.0,  0.0,  0.0],r'$\Gamma$']], [int(npoints*2),int(npoints),int(sqrt(5)*npoints)] )

# Load databases

# Electron energies database
elec = YamboElectronsDB.from_db_file(folder='./SAVE')

# Lattice information
lat  = YamboLatticeDB.from_db_file(filename='./SAVE/ns.db1')

# Exciton database read from db file
if os.path.isfile(bse_dir+'/ndb.BS_diago_Q01'):
    yexc = YamboExcitonDB.from_db_file(lat,filename='ndb.BS_diago_Q01',folder=bse_dir)
if os.path.isfile(bse_dir+'/ndb.BS_diago_Q1'):
    yexc = YamboExcitonDB.from_db_file(lat,filename='ndb.BS_diago_Q1',folder=bse_dir)

print("Ground state energy: %lf" % yexc.eigenvalues[0].real )
print("Intensity: %lf" % (yexc.get_intensities()[0].real+yexc.get_intensities()[1].real) )
print("1st-excited state energy: %lf" % yexc.eigenvalues[2].real )
print("Intensity: %lf" % (yexc.get_intensities()[2].real+yexc.get_intensities()[3].real) )

# List of states to be merged
states = [1,2]

# 1. Plot exciton weights in band structure NOT interpolated
exc_bands_up, exc_bands_dn = yexc.get_exciton_bs_spin_pol(elec,path,states,size_up=1.0, size_dw = 1.0)
exc_bands_up.plot_ax(ax,c_bands='red',c_weights='orange')
exc_bands_dn.plot_ax(ax,c_bands='blue',c_weights='orange')
ax.set_ylim(ymin,ymax)
plt.savefig('plot1.png')
plt.show()

# 2. Plot exciton weights in band structure INTERPOLATED

fig = plt.figure(figsize=(4,6))
ax  = fig.add_axes( [ 0.15, 0.15, 0.80, 0.80 ])

# In case of problems with the interpolation, try to increase lpratio
exc_bands_inter_up, exc_bands_inter_dn  = yexc.interpolate_spin_pol(elec,path,states,lpratio=10,f=None,size=0.5,verbose=True)
exc_bands_inter_up.plot_ax(ax,c_bands='red',c_weights='orange',alpha_weights=0.5,c_label='$X_1$')
exc_bands_inter_dn.plot_ax(ax,c_bands='blue',c_weights='orange',alpha_weights=0.5,c_label='$X_1$')
ax.set_ylim(ymin,ymax)
plt.savefig('plot2.png')
plt.show()


# 3. Plot exciton weights in a 2D map of the BZ

fig = plt.figure(figsize=(4,4))
ax  = fig.add_axes( [ 0.15, 0.15, 0.80, 0.80 ])

#yexc.plot_exciton_2D_ax(ax,states,mode='hexagon',limfactor=0.8,scale= 600)
yexc.plot_exciton_2D_ax(ax,states,limfactor=0.8,scale= 600)
plt.savefig('plot3.png')
plt.show()



from matplotlib.patches import RegularPolygon,Rectangle,Circle
from matplotlib.colors import to_rgba
from yambopy.lattice import bravais_types
import numpy as np

def add_fig_kwargs(func):
    """
    Decorator that adds keyword arguments for functions returning matplotlib
    figures.

    Taken from pymatgen:
    http://pymatgen.org/
    https://github.com/materialsproject/pymatgen
    """
    def wrapper(*args, **kwargs):
        # pop the kwds used by the decorator.
        title = kwargs.pop("title", None)
        show = kwargs.pop("show", True)
        savefig = kwargs.pop("savefig", None)

        # Call func and return immediately if None is returned.
        fig = func(*args, **kwargs)
        if fig is None:
            return fig

        # Operate on matplotlib figure.
        if title is not None:
            fig.suptitle(title)
        if savefig:
            fig.savefig(savefig)
        if show:
            import matplotlib.pyplot as plt
            plt.show()
        return fig
    return wrapper

def BZ_Wigner_Seitz(lattice,center=(0.,0.),orientation=np.radians(30),color='white',linewidth=2):
    """
    Wrapper function to decide which BZ shape to show 
    (will be a 2D slice if the lattice is 3D)

    Lattice types supported
    - hexagonal
    - square, rectangular, centered rectangular
    Lattice types unsupported
    - oblique

    """
    NoPatch = Circle((0,0),radius=0,visible=False)

    lat_type = bravais_types(lattice.lat,lattice.alat[0])
    if lat_type[:3]=='Hex': return BZ_hexagon(lattice.rlat,center=center,orientation=orientation,color=color,linewidth=linewidth)
    if lat_type[:3]=='Ort': return BZ_rectangle(lattice.rlat,color=color,linewidth=linewidth)

    if lat_type[:3]!='Hex' or lat_type[:3]!='Ort':
        print("[WARNING] Lattice type %s currently not supported for drawing BZ borders")
        return NoPatch

def BZ_hexagon(rlat,center=(0.,0.),orientation=np.radians(30),color='white',linewidth=2):
    """
    Returns hexagonal borders of 2D Wigner-Seitz cells to aid in k/q-space plotting
    to be added with ax.add_patch(BZ_hexagon)

    - rlat: reciprocal lattice vectors from YamboLatticeDB
    - center,orientation, color, linewidth are RegularPolygon parameters
    """

    # Reshape rlat
    rlat = np.array([[rlat[0,0],rlat[0,1]],[rlat[1,0],rlat[1,1]]])

    # Hexagon radius
    radius = np.linalg.norm(rlat[0]/np.sqrt(3.))

    # Matplotlib patch
    hexagon=RegularPolygon(center,numVertices=6,radius=radius,\
                           orientation=orientation,facecolor=to_rgba('white',0.),\
                           edgecolor=to_rgba(color,1.),linewidth=linewidth,zorder=10)

    return hexagon

def BZ_rectangle(rlat,color='white',linewidth=2):
    """
    Returns square borders of 2D Wigner-Seitz cells to aid in k/q-space plotting
    to be added with ax.add_patch(BZ_rectangle)

    - rlat: reciprocal lattice vectors from YamboLatticeDB
    - color, linewidth are RegularPolygon parameters
    """

    # Reshape rlat
    rlat = np.array([[rlat[0,0],rlat[0,1]],[rlat[1,0],rlat[1,1]]])

    # Matplotlib patch
    width, height  = rlat[0,0], rlat[1,1]
    origin = [-width/2.,-height/2.]
    rectangle = Rectangle(origin,width,height,facecolor=to_rgba('white',0.),\
                          edgecolor=to_rgba(color,1.),linewidth=linewidth,zorder=10)

    return rectangle

def shifted_grids_2D(k,b):
    """
    Shift a 2D k/q-point in the adjacent BZs.

    Inputs: 
    - k is kgrid in c.c.
    - b is ylat.rlat in c.c. (reciprocal lattice vectors) 

    Returns a list of the original plus the eight 2D shifted grids
    """
    N_k = len(k)
    N_dim = 2
    N_BZ  = 9
    
    c = []
    for i in range(-N_dim+1,N_dim): # Fill translation coefficients
        for j in range(-N_dim+1,N_dim): c.append( [i,j] )

    shifted_grids = []
    for ig in range(N_BZ): # Each shifted grid is an element of the list
        k_shift = np.zeros((N_k,3))
        for i in range(N_dim): 
            k_shift[:,i] = k[:,i]+c[ig][0]*b[0,i]+c[ig][1]*b[1,i]   
        shifted_grids.append(k_shift)

    return shifted_grids

def plot_mesh_2D_BZ(lattice,car_pts,car_pts2=None,save=False,**kwargs):
    """
    Fast plot of a k- or q-mesh in the 2D BZ with
    annotated indices and in CARTESIAN coordinates. 
    Supports also a second mesh for comparisons.
    
    This function is intended to help with debug, tests,
    developments.
    """
    import matplotlib.pyplot as plt
    marker = kwargs.pop('marker','H')
    size   = kwargs.pop('size',200)
    color  = kwargs.pop('color','teal')
    lwidth = kwargs.pop('lwidth',0.5)
    ecolor = kwargs.pop('ecolor','black')
    label  = kwargs.pop('label','grid 1')
    offset_xy = [kwargs.pop('offset_x',0.003),kwargs.pop('offset_y',0.005)]
    marker2 = kwargs.pop('marker2','h')
    size2   = kwargs.pop('size2',100)
    color2  = kwargs.pop('color2','orange')
    lwidth2 = kwargs.pop('lwidth2',0.5)
    ecolor2 = kwargs.pop('ecolor2','black')
    label2  = kwargs.pop('label2','grid 2')
    offset_xy2 = [kwargs.pop('offset_x2',-0.005),kwargs.pop('offset_y2',0.005)]

    # Do a 2D scatterplot of the kpoints in Cartesian coordinates
    fig = plt.figure(figsize=(9,9))
    ax = plt.gca()

    ## Add BZ borders
    ax.add_patch(BZ_Wigner_Seitz(lattice,color='black',linewidth=1.))

    ## Plot with "nice" layout
    ax.set_aspect('equal')
    ax.scatter(car_pts[:,0],car_pts[:,1],marker=marker,s=size,color=color,\
                linewidth=lwidth,edgecolors=ecolor,label=label)

    ## Explicitly show kpt indices
    for i_k,kpt in enumerate(car_pts):
        kx,ky = kpt[0],kpt[1]
        ax.annotate(i_k, (kx,ky), color=color, xytext=(kx+offset_xy[0],ky+offset_xy[1]))

    ## Plot second mesh for comparison if needed
    if car_pts2 is not None:
        ax.scatter(car_pts2[:,0],car_pts2[:,1],marker=marker2,s=size2,color=color2,\
                    linewidth=lwidth2,edgecolors=ecolor2,label=label2)
        for i_k,kpt in enumerate(car_pts2):
            kx,ky = kpt[0],kpt[1]
            ax.annotate(i_k, (kx,ky), color=color2, xytext=(kx+offset_xy2[0],ky+offset_xy2[1]))

    plt.legend()
    if not save: plt.show()
    else: plt.savefig('grids.png')



# Copyright (c) 2022
# All rights reserved.
#
# This file is part of the yambopy project
#
import os
import numpy as np
from netCDF4 import Dataset
from yambopy.tools.string import marquee

class YamboBZgridsDB(object):
    """
    This class reads the database ndb.kindx. 
    It is mainly used to maintain consistency in the indices of the k/q-point
    expansions from IBZ to full BZ between yambo and yambopy.
    
    Actual kpoint coordinates are found as in YamboLatticeDB attributes.
    Actual qpoint coordinates are found as YamboEm1sDB, YamboElphonDB, YamboExcitonDB, YamboExcitonPhononDB attributes.
    
    For reference check src/bz_ops/bz_samp_indexes.F in the yambo source.
    NB: The -1 are needed only to match python and fortran indices in python.
    
    ! ikbz=(ik,is) --<--:--<-- okbz=(ok,os) = (IK-Q)
    !                   :
    !                  /:\ iqbz=(iq,is)
    !                   :
    !
    ! iq_is = ik_is-ok_os-Go
    !
    ! qindx_X(iq,ikbz,1)-1=okbz
    ! qindx_X(iq,ikbz,2)-1=iGo
    !
    ! qindx_B(okbz,ikbz,1)-1=iqbz
    ! qindx_B(okbz,ikbz,2)-1=iGo
    !
    ! qindx_S(ik,iqbz,1)-1=okbz
    ! qindx_S(ik,iqbz,2)-1=iGo
    !
    ! qindx_C(ikbz,iqbz,1)-1=okbz
    ! qindx_C(ikbz,iqbz,2)-1=iGo    

    NB: the indices included in the tables are in fortran ordering (starting from 1)
    """
    def __init__(self,filename):
            
        if not os.path.isfile(filename):
            raise FileNotFoundError("error opening %s in YamboKptsDB"%filename)

        with Dataset(filename) as database:

            if 'CH_GRIDS' in database.variables: grid_str = 'CH_GRIDS'
            elif 'GRIDS_CH' in database.variables: grid_str = 'GRIDS_CH'
            self.grid_types = database.variables[grid_str][0].tobytes().decode('utf-8').strip()
            
            if 'X' in self.grid_types: self.qindx_X =  database.variables['Qindx'][:].T
            if 'B' in self.grid_types: self.qindx_B =  database.variables['Bindx'][:].T
            if 'S' in self.grid_types: self.qindx_S =  database.variables['Sindx'][:].T
            if 'C' in self.grid_types: self.qindx_C =  database.variables['Cindx'][:].T

    def get_string(self,mark="="):
        lines = []; app = lines.append
        app( marquee(self.__class__.__name__,mark=mark) )
        app( "Tables:                   %s"%self.grid_types)
        if 'S' in self.grid_types:
            app( "number of kpoints (IBZ):  %s"%self.qindx_S.shape[0])
        if 'X' in self.grid_types:
            app( "number of kpoints (BZ):   %s"%self.qindx_X.shape[1])
            app( "number of qpoints (IBZ):  %s"%self.qindx_X.shape[0])    
        if 'S' in self.grid_types:
            app( "number of qpoints (BZ):   %s"%self.qindx_S.shape[1])
        return '\n'.join(lines)
    
    def __str__(self):
        return self.get_string()



# Copyright (C) 2018 Henrique Pereira Coutada Miranda, Alejandro Molina-Sanchez
# All rights reserved.
#
# This file is part of yambopy
#
#
from __future__ import print_function
import os
import subprocess
from textwrap import dedent
from copy import deepcopy
from collections import OrderedDict
from .scheduler import Scheduler

class Pbs(Scheduler):
    """
    Class to submit jobs through the PBS scheduler.

    ``_vardict`` states the default assignement of the nodes and cores variables
    from the schduler class to the variables needed in this class
    """
    _vardict = {"cores":"core",
                "nodes":"select"}
                   
    def initialize(self):
        self.get_vardict()
        args = self.arguments
        queue = self.get_arg("queue")
        rerunable= self.get_arg("rerunable")
        mem = self.get_mem()
        if self.name: args.append("-N %s"%self.name)
        
        if queue: args.append("-q %s"%(queue))
        group_list = self.get_arg("group_list")
        
        if group_list: args.append("-W group_list=%s"%group_list)
        dependent = self.get_arg("dependent")
        
        if dependent: args.append("-W depend=afterok:%s"%dependent)
        args.append("-l walltime=%s"%self.walltime)

        if rerunable: args.append("-r y")

        if self.get_arg("pvmem"): args.append("-l pvmem=%dMB"%mem)
        
        resources_line = self.get_resources_line()
        if resources_line:
            args.append("-l %s"%resources_line)

    def get_mem(self):
        """
        get the memory for this job
        """

        #block to evaluate expressions from
        #http://stackoverflow.com/questions/2371436/evaluating-a-mathematical-expression-in-a-string
        import ast
        import operator as op

        # supported operators
        operators = {ast.Add: op.add, ast.Sub: op.sub, ast.Mult: op.mul,
                            ast.Div: op.truediv, ast.Pow: op.pow, ast.BitXor: op.xor,
                                         ast.USub: op.neg}

        def eval_expr(expr):
            return eval_(ast.parse(expr, mode='eval').body)

        def eval_(node):
            if isinstance(node, ast.Num): # <number>
                return node.n
            elif isinstance(node, ast.BinOp): # <left> <operator> <right>
                return operators[type(node.op)](eval_(node.left), eval_(node.right))
            elif isinstance(node, ast.UnaryOp): # <operator> <operand> e.g., -1
                return operators[type(node.op)](eval_(node.operand))
            else:
                raise TypeError(node)
        ######

        mem = self.get_arg("mem")
        if mem:
            if self.cores: cores = self.cores
            else: cores = 1
            if self.nodes: nodes = self.nodes
            else: nodes = 1
            mem = mem.replace("nodes",str(nodes))
            mem = mem.replace("cores",str(cores))
            mem = eval_expr(mem) 
        return mem 

    def get_resources_line(self):
        """
        get the the line with the resources
        """
        tags = ['select','nodes','core','ppn','ncpus','mpiprocs','ompthreads']
        args = [self.get_arg(tag) for tag in tags]
        resources = []
        if self.nodes: resources.append((self.vardict['nodes'],self.nodes))
        if self.cores: resources.append((self.vardict['cores'],self.cores))
        resources += [(tag,value) for tag,value in zip(tags,args) if value is not None]
        resources = OrderedDict(resources)

        # memory stuff
        mem = self.get_mem()
        if mem: resources["mem"]  = "%dMB"%mem
        
        resources_line = ":".join(["%s=%s"%(item,value) for item,value in list(resources.items())])
       
        return resources_line
    
    def get_script(self):
        """
        get a .pbs file to be submitted using qsub
        qsub <filename>.pbs
        """
        s = '#!/bin/bash\n'
        s += "\n".join(["#PBS %s"%s for s in self.arguments])+'\n'
        s += self.get_commands()
        return s
        
    def get_bash(self):
        """
        get a bash command to submit the job
        """
        s = "echo \"%s\" | "%self.get_commands().replace("'","\"").replace("\"","\\\"").replace("$","\$")
        s += "qsub \\\n"
        s += " \\\n".join(self.arguments)
        return s
        
    def __str__(self):
        return self.get_script()
        
    def run(self,filename='run.sh',dry=False,command="qsub",verbose=0):
        """
        run the command
        arguments:
        dry - only print the commands to be run on the screen
        """
        if dry:
            print(self)
            return

        #create the submission script
        self.write(filename)
        workdir  = os.path.dirname(filename)
        basename = os.path.basename(filename) 

        p = subprocess.Popen([command,basename],stdout=subprocess.PIPE,stderr=subprocess.PIPE,cwd=workdir)
        self.stdout,self.stderr = p.communicate()
        
        #check if there is stderr
        if self.stderr: raise Exception(self.stderr)
        
        #check if there is stdout
        if verbose: print(self.stdout)
        
        #get jobid
        self.jobid = str(self.stdout).split("\n")[0]
        if verbose: print("jobid:",self.jobid)

    



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: FP
#
# This file is part of the yambopy project
#
from yambopy.common.save_generation import CreateYamboSave
from yambopy.common.calculation_manager import check_qe_completed, shell_qe_run
from yambopy.io.iofile import YamboIO
from schedulerpy import Scheduler
import os
from copy import deepcopy

class YamboGkkpCompute():
    """
    Class to obtain qe s.dbph* and yambo ndb.elph* databases starting from scratch.

    [NOTE: this class uses the old el-ph yambo interface (via double ph.x calculation + ypp_ph)]
    [      A new interface making use of the LetzElphC code is available                       ]
    
    It runs the necessary pw.x and ph.x simulations, optionally followed by the yambo setup.
    
    Inputs needed:
        <required> 
        - scf_input: pw scf input file [NOTE: dvscf and gkkp inputs are automatically generated, check parameters if interested]

        <optional>
        - work_dir: directory where flow is run and yambo SAVE will appear
        - pw_exec_path: path to executables
        - qe_scheduler: optional scheduler for cluster submission
        - with_SAVE: if True, workflow will generate yambo SAVE at the end (the python master process will remain active).
                     The workflow can be called a second time switching with_SAVE to True to immediately generate the SAVE.

          [SUBOPTIONS for with_SAVE]

              -- expand: whether to expand the matrix elements
              -- yambo_exec_path: path to executables

    TODO: allow for random qpoints in ph calculations
    """
    
    def __init__(self,scf_input,work_dir='.',pw_exec_path='',qe_scheduler=None,with_SAVE=False,yambo_exec_path='',expand=False):

        if not os.path.isdir(work_dir): os.mkdir(work_dir)
        self.RUN_path = os.path.abspath(work_dir)
        self.wait_up = with_SAVE #Slightly restructure dependencies and waith for job completions if SAVE is to be created

        #Configuring schedulers
        if qe_scheduler is not None: self.qejobrun= qe_scheduler                          # Here we use, e.g., slurm 
        else:                        self.qejobrun = Scheduler.factory(scheduler="bash")  # Run without submission

        #Executables
        if yambo_exec_path != '': yambo_exec_path+='/'
        self.yambo     = yambo_exec_path + 'yambo'
        self.yambo_ph  = yambo_exec_path + 'yambo_ph' 
        self.p2y       = yambo_exec_path + 'p2y'
        self.ypp_ph    = yambo_exec_path + 'ypp_ph'
        if pw_exec_path != '': pw_exec_path+='/'
        self.pw = pw_exec_path + 'pw.x'
        self.ph = pw_exec_path + 'ph.x'
        self.dynmat = pw_exec_path + 'dynmat.x'

        # Inputs
        self.scf_input = scf_input
        self.prefix    = scf_input.prefix
        
        # Output names
        self.out_scf   = 'scf.out'
        self.out_nscf  = 'nscf.out'
        self.out_dvscf = 'dvscf.out'
        self.out_gkkp  = 'gkkp.out'
        
        #Start IO
        self.yf = YamboIO(out_name='YAMBOPY_gkkp_calculation.log',out_path=self.RUN_path,print_to_shell=True)
        self.yf.IO_start()
        
        self.yf.msg('#### GKKP WORKFLOW ####')

        # Create folder structure
        self.setup_calculations()

        # Run jobs
        if not self.scf_status:
            self.yf.msg('Running scf.')
            self.run_scf()
        
        if not self.dvscf_status:
            self.yf.msg('Running dvscf.')
            self.run_dvscf()

        if not self.gkkp_status:
            self.yf.msg('Running gkkp.')
            self.run_gkkp()
        
        if not self.nscf_status:
            self.yf.msg('Running nscf.')
            self.run_nscf()
            
        # [OPTIONAL] Create SAVE
        if with_SAVE:
            self.setup_SAVE()

            if not self.are_gkkp_there:
                if self.is_SAVE_there: 
                     import shutil
                     shutil.rmtree('%s/SAVE'%self.RUN_path)
                if expand: save_type = 'expanded_elph'
                else:      save_type = 'elph'
            
                self.yf.msg('---- Generating SAVE folder: ----') 
                CreateYamboSave(self.prefix,save_type=save_type,nscf=self.nscf_dir,elph_path=self.gkkp_dir,database=self.RUN_path,\
                                yambo_exec_path=yambo_exec_path,printIO=True)
                self.clean_rubbish()

        #End IO        
        self.yf.IO_close()
        
    def setup_calculations(self):
        """
        Generate workflow tree
        """
        
        # Directory names (hardcoded)
        dft_dir  = '%s/dft'%self.RUN_path
        self.scf_dir  = '%s/scf'%dft_dir
        self.gkkp_dir = '%s/gkkp'%dft_dir
        self.nscf_dir = '%s/nscf'%dft_dir
        
        # Logicals
        self.gkkp_status  = False
        self.dvscf_status = False
        self.scf_status   = False
        self.nscf_status  = False
        
        if not os.path.isdir(dft_dir):  os.mkdir(dft_dir)
        if not os.path.isdir(self.scf_dir):  os.mkdir(self.scf_dir)
        if not os.path.isdir(self.gkkp_dir): os.mkdir(self.gkkp_dir)
        if not os.path.isdir(self.nscf_dir): os.mkdir(self.nscf_dir)
        
        # Check if any gkkp->dvscf->scf calculations have been already done
        self.gkkp_status = check_qe_completed(self.gkkp_dir,self.prefix,self.out_gkkp,calc_type='gkkp')
        if self.gkkp_status: 
            self.yf.msg('gkkp calculation found!') 
            self.dvscf_status = True
        else:
            self.dvscf_status = check_qe_completed(self.gkkp_dir,self.prefix,self.out_dvscf,calc_type='ph')
            if self.dvscf_status: 
                self.yf.msg('dvscf calculation found!')
            else:
                self.scf_status = check_qe_completed(self.scf_dir,self.prefix,self.out_scf,calc_type='pw')
                if self.scf_status: 
                    self.yf.msg('scf calculation found!')
                       
        # Check if any nscf->scf calculations have been already done
        self.nscf_status = check_qe_completed(self.nscf_dir,self.prefix,self.out_nscf,calc_type='pw')
        if self.nscf_status: 
             self.yf.msg('nscf calculation found')
             self.scf_status = True
        else:
             if not self.scf_status:
                  self.scf_status = check_qe_completed(self.scf_dir,self.prefix,self.out_scf,calc_type='pw')
                  if self.scf_status: self.yf.msg('scf calculation found!')

    def setup_SAVE(self):
        """
        Expand the workflow tree to include yambo SAVE
        """

        # Check if SAVE and/or gkkp dbs are there already
        save_dir = '%s/SAVE'%self.RUN_path
        if os.path.isdir(save_dir):
            self.yf.msg('SAVE folder found!')
            self.is_SAVE_there = True
            if os.path.isfile('%s/ndb.elph_gkkp'%save_dir) or os.path.isfile('%s/ndb.elph_gkkp_expanded'%save_dir):
                self.yf.msg('ndb.elph databases already found!')
                self.are_gkkp_there = True
            else:
                self.are_gkkp_there = False
        else:
            self.is_SAVE_there  = False
            self.are_gkkp_there = False
            
    def run_scf(self):
        """
        Run scf calculation
        """
        if self.scf_input.system['nbnd'] is None:
            raise ValueError('Please specify nbnd in the scf input in order to be able to compute the gkkp elements.')
        
        # Write down input
        inp_name = self.prefix + '.scf'
        self.scf_input.write('%s/%s'%(self.scf_dir,inp_name))
        
        # Submit calculation
        jname = 'scf'
        self.scf_id = shell_qe_run(jname,inp_name,self.out_scf,self.scf_dir,exec=self.pw,scheduler=self.qejobrun)
        
    def run_dvscf(self):
        """
        Run dvscf calculation
        """
        # Generate and write down input
        dvscf_input = self.generate_ph_input('dvscf')
        inp_name = self.prefix + '.dvscf'
        dvscf_input.write('%s/%s'%(self.gkkp_dir,inp_name))
        
        # Generate and write down dynmat input
        dynmat_input = self.generate_dynmat_input()
        dynp_name = self.prefix + '.dynmat'
        dynmat_input.write('%s/%s'%(self.gkkp_dir,dynp_name))

        # Set dynmat run after completion of main task
        dyn_run = ["mpirun -np 1 %s -inp %s > dynmat.out"%(self.dynmat,dynp_name)]

        # Create symlink to qe save if needed
        commands = []
        if not os.path.islink('%s/%s.save'%(self.gkkp_dir,self.prefix)):
            commands.append('ln -s %s/%s.save %s/'%(self.scf_dir,self.prefix,self.gkkp_dir)) 

        # Manage dependency
        if self.scf_status: depend = None # No dependency if scf was found
        else: depend = self.scf_id
        
        # Submit calculation
        jname = 'dvscf'
        self.dvscf_id = shell_qe_run(jname,inp_name,self.out_dvscf,self.gkkp_dir,exec=self.ph,shell_name='dvscf',\
                                     scheduler=self.qejobrun,pre_run=commands,pos_run=dyn_run,depend_on_JOBID=depend) 
                                     
    def run_gkkp(self):
        """
        Run gkkp calculation
        """
        # Generate and write down input
        gkkp_input = self.generate_ph_input('gkkp')
        inp_name = self.prefix + '.gkkp'
        gkkp_input.write('%s/%s'%(self.gkkp_dir,inp_name))
        
        # Create symlink to qe save if needed
        commands = []
        if not os.path.islink('%s/%s.save'%(self.gkkp_dir,self.prefix)):
            commands.append('ln -s %s/%s.save %s/'%(self.scf_dir,self.prefix,self.gkkp_dir)) 
        
        # Manage dependency
        if self.dvscf_status: depend = None # No dependency if dvscf was found
        else: depend = self.dvscf_id

        # Submit calculation
        jname = 'gkkp'
        self.gkkp_id = shell_qe_run(jname,inp_name,self.out_gkkp,self.gkkp_dir,exec=self.ph,shell_name='gkkp',\
                                    scheduler=self.qejobrun,pre_run=commands,depend_on_JOBID=depend)    
                                    
    def run_nscf(self):
        """
        Run nscf calculation
        """
        # Generate and write down input
        nscf_input = self.generate_nscf_input()
        inp_name = self.prefix + '.nscf'
        nscf_input.write('%s/%s'%(self.nscf_dir,inp_name))
        
        # Create symlink to qe save if needed
        commands = []
        if not os.path.isdir('%s/%s.save'%(self.nscf_dir,self.prefix)):
            commands.append('cp -r %s/%s.save %s/'%(self.scf_dir,self.prefix,self.nscf_dir)) 
                
        # Dependency here may include gkkp job to ensure that this is the last job to be completed if SAVE is to be generated
        if self.wait_up:
            if self.gkkp_status and self.scf_status:       depend = None # No dependency if scf and gkkp were found
            elif self.gkkp_status and not self.scf_status: depend = self.gkkp_id # scf was found, not gkkp
            elif not self.gkkp_status and self.scf_status: depend = self.scf_id  # gkkp was found, not scf
            else:                                          depend = '%s:%s'%(self.scf_id,self.gkkp_id) # double dependency
        else: 
            if self.scf_status: depend = None # No dependency if scf was found
            else:               depend = self.scf_id    
        
        # Submit calculation
        jname = 'nscf'
        self.nscf_id = shell_qe_run(jname,inp_name,self.out_nscf,self.nscf_dir,exec=self.pw,scheduler=self.qejobrun,\
                                    pre_run=commands,depend_on_JOBID=depend,hang_python=self.wait_up)       

    def generate_nscf_input(self):
        """
        Create nscf input for yambo SAVE starting from scf input
        """
        
        nscf_input = deepcopy(self.scf_input)
        nscf_input.control['calculation']="'nscf'"
        nscf_input.electrons['diago_full_acc'] = ".true."
        nscf_input.electrons['conv_thr'] = 1e-8
        nscf_input.system['force_symmorphic'] = ".true."
        
        return nscf_input

    def generate_ph_input(self,mode):
       """
       Create dvscf or gkkp input starting from scf input
 
       - mode: either 'dvscf' or 'gkkp'
       """

       from qepy import PhIn 
       ph_input = PhIn()
       # Common to dvscf and gkkp
       ph_input['prefix'] = "'%s'"%self.prefix
       ph_input['fildyn'] = "'%s'"%(self.prefix+'.dyn')
       nq1,nq2,nq3 = [ int(nk) for nk in self.scf_input.kpoints ]
       ph_input.set_nq(nq1,nq2,nq3)
       ph_input['tr2_ph'] = 1e-14
       ph_input['fildvscf']="'dvscf'"
       ph_input['ldisp']='.true.'
       ph_input['qplot']='.false.'
       # Only dvscf
       if mode=='dvscf':

           # Add effective charges if dealing with a non-metal
           is_insulator = 'occupations' not in self.scf_input.system or self.scf_input.system['occupations'] != "'smearing'"
           if is_insulator: ph_input['epsil']='.true.'
           self.is_insulator = is_insulator

           ph_input['electron_phonon']="'dvscf'"
           ph_input['recover']='.true.'
           ph_input['trans']='.true.'

       elif mode=='gkkp':
           ph_input['electron_phonon']="'yambo'"
           ph_input['trans']='.false.'
       else: raise ValueError("ph input mode not recognized (either 'dvscf' or 'gkkp')")

       return ph_input

    def generate_dynmat_input(self):
       """
       Create dynmat input file in order to treat issues with the frequencies at the Gamma point:
         -- Apply the acoustic sum rule
         -- Correct the LO mode (if non-metal)

         Outputs are saved in the gkkp folder as prefix.GAMMA_eigs_eivs and prefix.GAMMA_eigs_norm_eivs
       """
       from qepy import DynmatIn
       dm_input = DynmatIn()
       dm_input['asr']="'crystal'"
       dm_input['fildyn']="'%s.dyn1'"%self.prefix
       dm_input['fileig']="'%s.GAMMA_eigs_eivs'"%self.prefix
       dm_input['filout']="'%s.GAMMA_eigs_norm_eivs'"%self.prefix

       # Add LO correction along first cartesian axis if dealing with non-metal
       if self.is_insulator:
           dm_input['q(1)']=1
           dm_input['q(2)']=0
           dm_input['q(3)']=0

       return dm_input

    def clean_rubbish(self):
       """
       Remove logs, reports and inputs generated during SAVE creation
       """
       from glob import glob
       run_dir = self.RUN_path+'/'
       logs1 =    glob(run_dir+'l-*')
       logs2 =    glob(run_dir+'l_*')
       reports1 = glob(run_dir+'r-*')
       reports2 = glob(run_dir+'r_*')
       setups   = glob(run_dir+'setup.in*')
       for log in logs1:       os.remove(log)
       for log in logs2:       os.remove(log)
       for report in reports1: os.remove(report)
       for report in reports2: os.remove(report)
       for setup in setups:    os.remove(setup)
       os.remove(run_dir+'gkkp.in')
     



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: FP
#
# This file is part of the yambopy project
#
from yambopy import *
from netCDF4 import Dataset
from math import sqrt
import numpy as np
from yambopy.tools.string import marquee
import os
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from yambopy.units import ha2ev, ev2cm1, I
from yambopy.plot.plotting import add_fig_kwargs,BZ_Wigner_Seitz,shifted_grids_2D

class YamboElectronPhononDB():
    """
    Python class to read the electron-phonon matrix elements from yambo.
    
    By default it reads the full databases including fragments.
    
    - Input: YamboLatticeDB object, paths of ndb.elph* and ns.db1
    - Input: if not read_all, read only header
    
    - Usage and main variables: 
    
      :: yph = YamboElectronPhononDB(ylat,folder_gkkp=path1,save=path2)
    
      :: yph.ph_energies     #Phonon energies (eV)      
      :: yph.ph_eigenvectors #Phonon modes
      :: yph.gkkp            #El-ph matrix elements (by default normalised with ph. energies):
      :: yph.gkkp_sq         #Couplings (square) 

      Additional variables (Experimental stuff)
      :: yph.gkkp_bare
      :: yph.gkkp_bare_sq
      :: yph.gkkp_mixed      #Coupling (mixed bare-dressed)
   
    Formats:
    - modes[il][iat][ix]
    - gkkp[iq][ik][il][ib1][ib2]

    Plots provided:
    - Call function plot for scatterplot in the k-BZ of any quantity A(k)_{iq,ib1,ib2,inu} 
        -- if plt_show, show plot at runtime
        -- if plt_cbar, add colorbar

      Example, plot of |g(k)_{0,3,4,4}|:      
           :: yph.plot_elph( np.abs(yph.gkkp[0,:,3,4,4]) )              
    """
    def __init__(self,lattice,filename='ndb.elph_gkkp',folder_gkkp='SAVE',save='SAVE',read_all=True):
        
        self.lattice = lattice

        # Find correct database names
        if os.path.isfile("%s/ndb.elph_gkkp"%folder_gkkp): filename='%s/ndb.elph_gkkp'%folder_gkkp
        elif os.path.isfile("%s/ndb.elph_gkkp_expanded"%folder_gkkp): filename='%s/ndb.elph_gkkp_expanded'%folder_gkkp
        else: filename = "%s/%s"%(folder_gkkp,filename)
        self.frag_filename = filename + "_fragment_"
        self.are_bare_there = False
        
        # necessary lattice information
        self.alat        = lattice.alat
        self.rlat        = lattice.rlat
        self.car_kpoints = lattice.car_kpoints
            
        # Check if databases exist. Exit only if header is missing.
        try: database = Dataset(filename)
        except: raise FileNotFoundError("error opening %s in YamboElectronPhononDB"%filename)
        
        try: database_frag = Dataset("%s1"%self.frag_filename)
        except FileNotFoundError: print("[WARNING] Database fragment at q=0 not detected")
        else: 
            # Check if bare matrix elements are present
            try: 
                database_frag.variables['ELPH_GKKP_BARE_Q1']
            except KeyError: 
                database_frag.close()
            else:
                self.are_bare_there = True
                database_frag.close()
                 
        #read qpoints    
        self.qpoints = database.variables['PH_Q'][:].T
        self.car_qpoints = np.array([ q/self.alat for q in self.qpoints ])
        #read dimensions of electron phonon parameters
        self.nmodes, self.nqpoints, self.nkpoints, b_1, b_2 = database.variables['PARS'][:5].astype(int)
        if b_1>b_2: # Old database (no GkkpBands in PARS)
            self.nbands = b_1
            self.b_in, self.b_out = [0,self.nbands-1]
        else: # New database (PARS with GkkpBands)
            self.b_in, self.b_out = [b_1-1,b_2-1]
            self.nbands = b_2-b_1+1
        self.natoms = int(self.nmodes/3)

        try: #check if IBZ K-point list is provided
            self.ibz_kpoints_elph = database.variables['HEAD_KPT'][:].T
            self.ibz_car_kpoints = np.array([ k/self.alat for k in self.ibz_kpoints_elph ])
        except KeyError: 
            pass

        try: # Check if full K-point list is provided (upon expansion), otherwise use the one from ns.db1
            self.kpoints_elph = database.variables['PH_K'][:].T
            self.car_kpoints = np.array([ k/self.alat for k in self.kpoints_elph ])
            database.close()
        except KeyError:
            database.close()
        
        #Check how many databases are present
        self.nfrags = self.nqpoints
        for iq in range(self.nqpoints):
            if not os.path.isfile("%s%d"%(self.frag_filename,iq+1)): 
                self.nfrags = iq
                break
        
        # Keep reading
        if read_all: self.read_full_DB()
        
    def read_full_DB(self):
        """
        Read all variables in the ndb.elph_gkkp* dbs as attributes of this class
        """
        
        # Frequencies
        self.read_frequencies()
        
        # Eigenmodes
        self.read_eigenmodes()

        # <dVscf> matrix elements plus <dVbare> if they exist
        self.read_elph()
        if self.are_bare_there: self.read_elph(kind='bare')   
   
        # Get square matrix elements
        self.get_gkkp_sq()
        
        # Get the symmetrised dressed-bare coupling
        if self.are_bare_there: self.get_gkkp_mixed()

    def read_frequencies(self):
        """
        Read phonon frequencies in eV
        """
        self.ph_energies  = np.zeros([self.nfrags,self.nmodes])
        
        for iq in range(self.nfrags):
            fil = self.frag_filename + "%d"%(iq+1)
            database = Dataset(fil)
            self.ph_energies[iq] = np.sqrt(database.variables['PH_FREQS%d'%(iq+1)][:])*ha2ev
            database.close()
        
    def read_eigenmodes(self):
        """
        Read phonon eigenmodes
        """
        self.ph_eigenvectors = np.zeros([self.nfrags,self.nmodes,self.natoms,3],dtype=np.complex64)
        
        for iq in range(self.nfrags):
            fil = self.frag_filename + "%d"%(iq+1)
            database = Dataset(fil)
            #eigs_q[cartesian][atom][mode][complex]
            eigs_q = database.variables['POLARIZATION_VECTORS'][:].T
            self.ph_eigenvectors[iq] = eigs_q[0,:,:,:] + eigs_q[1,:,:,:]*I
            database.close()
             
    def read_elph(self,kind='dressed',scale_g_with_ph_energies=True):
        """
        Read electron-phonon matrix elements
        
        - kind is 'dressed' or 'bare'
        - var_nm is 'ELPH_GKKP_Q' or 'ELPH_GKKP_BARE_Q'
        - If scale_g_with_ph_energies they are divided by sqrt(2*ph_E)

        NB: ELPH_GKKP_Q is saved by yambo as (2,mode,bnd1,bnd2,k), but netCDF stores
            the *transpose* (k,bnd2,bnd1,mode,2).
            We want to change it to complex (k,mode,bnd1,bnd2), therefore we need to
            *swap* bnd2<->mode.
        """    
        if kind!='dressed' and kind!='bare': 
            raise ValueError("Wrong kind %s (can be 'dressed' [Default] or 'bare')"%kind) 
        if kind=='dressed': var_nm = 'ELPH_GKKP_Q'
        if kind=='bare':    var_nm = 'ELPH_GKKP_BARE_Q'       
        
        # gkkp[q][k][mode][bnd1][bnd2]
        gkkp_full = np.zeros([self.nfrags,self.nkpoints,self.nmodes,self.nbands,self.nbands],dtype=np.complex64)   
        
        for iq in range(self.nfrags):
            fil = self.frag_filename + "%d"%(iq+1)
            database = Dataset(fil)
            gkkp = database.variables['%s%d'%(var_nm,iq+1)][:]
            gkkp_full[iq] = np.swapaxes(gkkp[:,:,:,:,0] + I*gkkp[:,:,:,:,1],-1,1)
            database.close()
        
        # Check integrity of elph values
        if np.isnan(gkkp_full).any(): print('[WARNING] NaN values detected in elph database.')
        
        # Scaling with phonon energies
        if scale_g_with_ph_energies: gkkp_full = self.scale_g(gkkp_full)    
                
        if kind=='dressed': self.gkkp = gkkp_full
        if kind=='bare': self.gkkp_bare = gkkp_full          
    
    def scale_g(self,dvscf):
        """
        Normalise matrix elements by the phonon energy as: 
       
        g_qnu = dvscf_qnu/sqrt(2*w_qnu)
        """
        
        g = np.zeros([self.nfrags,self.nkpoints,self.nmodes,self.nbands,self.nbands],dtype=np.complex64)
        for iq in range(self.nfrags):
            for inu in range(self.nmodes): 
                if iq==0 and inu in [0,1,2]: 
                    g[iq,:,inu,:,:] = 0. # Remove acoustic branches
                else:
                    ph_E = self.ph_energies[iq,inu]/ha2ev # Put back the energies in Hartree units
                    g[iq,:,inu,:,:] = dvscf[iq,:,inu,:,:]/np.sqrt(2.*ph_E)
        return g

    def get_gkkp_sq(self,read_bare=False):
        """
        Return g^2
        """
        self.gkkp_sq = np.abs(self.gkkp)**2. 
        if self.are_bare_there: self.gkkp_bare_sq = np.abs(self.gkkp_bare)**2. 

    def get_gkkp_mixed(self):
        """
        Return the symmetrised dressed-bare coupling
        """
        self.gkkp_mixed = np.real(self.gkkp)*np.real(self.gkkp_bare)+np.imag(self.gkkp)*np.imag(self.gkkp_bare)
        
    @add_fig_kwargs
    def plot_elph(self,data,kcoords=None,plt_show=False,plt_cbar=False,**kwargs):
        """
        2D scatterplot in the BZ:

         (i)  in k-space of the quantity A_{k}(iq,inu,ib1,ib2).
         (ii) in q-space of the quantity A_{q}(ik,inu,ib1,ib2).       

        Any real quantity which is a function of only the k-grid or q-grid may be supplied.
        The indices iq/ik,inu,ib1,ib2 are user-specified.

        - kcoords refers to the k/q-grid in Cartesian coordinates (i.e., yelph.car_qpoints and similar).
          If None is specified, a k-space, fixed-q plot is assumed.
        
        - if plt_show plot is shown
        - if plt_cbar colorbar is shown
        - kwargs example: marker='H', s=300, cmap='viridis', etc.
        
        NB: So far requires a 2D system. 
            Can be improved to plot BZ planes at constant k_z for 3D systems.
        """        
        if kcoords is None: kpts = self.car_kpoints # Assume k-space plot
        else:               kpts = kcoords # Plot on momentum map supplied by user       

        # Input check
        if len(data)!=len(kpts): 
            raise ValueError('Something wrong in data dimensions (%d data vs %d kpts)'%(len(data),len(kpts)))
        
        # Global plot stuff
        self.fig, self.ax = plt.subplots(1, 1)
        self.ax.add_patch(BZ_Wigner_Seitz(self.lattice))
        
        if plt_cbar:
            if 'cmap' in kwargs.keys(): color_map = plt.get_cmap(kwargs['cmap'])
            else:                       color_map = plt.get_cmap('viridis') 
        lim = 1.05*np.linalg.norm(self.rlat[0])
        self.ax.set_xlim(-lim,lim)
        self.ax.set_ylim(-lim,lim)

        # Reproduce plot also in adjacent BZs
        BZs = shifted_grids_2D(kpts,self.rlat)
        for kpts_s in BZs: plot=self.ax.scatter(kpts_s[:,0],kpts_s[:,1],c=data,**kwargs)
        
        if plt_cbar: self.fig.colorbar(plot)
        
        plt.gca().set_aspect('equal')

        if plt_show: plt.show()
        else: print("Plot ready.\nYou can customise adding savefig, title, labels, text, show, etc...")
        
    def __str__(self,verbose=False):

        try: self.ph_energies
        except AttributeError: self.read_frequencies()

        try: self.ph_eigenvectors
        except AttributeError: self.read_eigenmodes()

        lines = []; app = lines.append
        app(marquee(self.__class__.__name__))
            
        app('nqpoints: %d'%self.nqpoints)
        app('nkpoints: %d'%self.nkpoints)
        app('nmodes: %d'%self.nmodes)
        app('natoms: %d'%self.natoms)
        app('nbands: %d (%d - %d)'%(self.nbands,self.b_in,self.b_out))
        if self.nfrags == self.nqpoints: app('fragments: %d'%self.nfrags)
        else: app('fragments: %d [WARNING] nfrags < nqpoints'%self.nfrags)
        if self.are_bare_there: app('bare couplings are present')
        if verbose:
            app('-----------------------------------')
            for iq in range(self.nfrags):
                app('nqpoint %d'%iq)
                for n,mode in enumerate(self.ph_eigenvectors[iq]):
                    app('mode %d freq: %lf meV'%(n,self.ph_energies[iq,n]*1000.))
                    for a in range(self.natoms):
                        app(("%12.8lf "*3)%tuple(mode[a].real))
            app('-----------------------------------')
        return "\n".join(lines)



# Copyright (C) 2018 Henrique Pereira Coutada Miranda, Alejandro Molina-Sanchez
# All rights reserved.
#
# This file is part of yambopy
#
from __future__ import print_function
import os
from builtins import str
import subprocess
import sys
from .scheduler import Scheduler

class Bash(Scheduler):
    """
    Class to submit jobs using BASH
    """
    _vardict = {"cores":"core",
                "nodes":"nodes"}

    def initialize(self):
        self.get_vardict()

    def __str__(self):
        return self.get_commands()

    def get_bash(self):
        return str(self)

    def get_script(self):
        return str(self)

    def add_mpirun_command(self, cmd):
        threads = 1
        if self.cores: threads*=self.cores
        mpirun = self.get_arg("mpirun","mpirun")
        np = self.get_arg("np","-np")
        self.add_command("%s %s %d %s"%(mpirun,np,threads,cmd))

    def run(self,filename='./run.sh',command='sh',dry=False):
        #create the submission script
        self.write(filename)
        workdir  = os.path.dirname(filename)
        basename = os.path.basename(filename)

        if dry:
            print(command)
        else:
            p = subprocess.Popen([command,basename],stdout=subprocess.PIPE,stderr=subprocess.PIPE,cwd=workdir)
            self.stdout, self.stderr = p.communicate()
            # In Python 3, Popen.communicate() returns bytes
            try:
                self.stdout = self.stdout.decode()
                self.stderr = self.stderr.decode()
            # If Python 2, <str>.decode() will raise an error that we ignore
            except AttributeError:
                pass



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: HPC
#
# This file is part of the yambopy project
#
import numpy as np
from yambopy.units import kb

def abs2(x):
    return x.real**2 + x.imag**2
 
def lorentzian(x,x0,g):
    height=1./(np.pi*g)
    return height*(g**2)/((x-x0)**2+g**2)

def gaussian(x,x0,s,max_exp=50.,min_exp=-100.):
    height=1./(np.sqrt(2.*np.pi)*s)
    argument=-0.5*((x-x0)/s)**2
    #Avoiding undeflow errors...
    np.place(argument,argument<min_exp,min_exp)
    return height*np.exp(argument)

def boltzman_f(Eb, Bose_Temp):
    return np.exp(-Eb/(kb*Bose_Temp))

def fermi(e,max_exp=50,min_exp=-100):
    """ fermi dirac function
    """
    if e > max_exp:
        return 0
    elif e < min_exp:
        return 1
    return 1/(np.exp(e)+1)

def fermi_array(e_array,ef,invsmear):
    """
    Fermi dirac function for an array
    """
    e_array = (e_array-ef)/invsmear
    return [ fermi(e) for e in e_array]

def bose(Eb,Bose_Temp,max_exp=50,thr=1e-10):
    """ 
    Bose-Einstein function (accepts ndarray)

    Eb --> in eV (should be positive)
    Bose_Temp --> in K

    If Eb=0 (e.g. phonon acoustic modes at q=0) it returns zero to avoid
    divide by zero errors: these states should be excluded from loops/calculations!
    """
    if Bose_Temp < thr: return np.zeros(Eb.shape) # zero temperature: no occupation
    e = Eb/(kb*Bose_Temp)
    # Ignore overflow div by zero and return zero if energy is zero 
    # (e.g. acoustic modes at q=0). 
    # THIS DOES NOT REPLACE HANDLING THE ZERO CASE EXPLICITLY IN YOUR APPLICATION
    with np.errstate(over='ignore',divide='ignore', invalid='ignore'): 
        n_be = np.where(e==0., 0.0, 1.0/(np.exp(e)-1.0))
    return n_be



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: HPC, FP
#
# This file is part of the yambopy project
#
"""
Create, read and write yambo input files
Read, modify and write yambo databases
Analyse results from yambo calculations

Modules:
    io
        - YamboIn: read, write and manipulate yambo input files
        - YamboOut: read yambo output files and save in .json
    dbs
        - YamboSaveDB: read information in the ns.db1
        - YamboLatticeDB: read lattice parameters, symmetries and k-points from ns.db1
        - YamboDipolesDB: dipole matrix elements from ndb.dip*
        - YamboStaticScreeningDB: static dielectric screening from ndb.em1s*
        - YamboElectronsDB: read the electronic states from ns.db1
        - YamboQPDB: read the quasiparticle energies db ndb.QP
        - YamboGreenDB: read the green's functions calculated using yambo
        - YamboExcitonsDB: read excitonic properties from a BSE calculation
        - YamboKernelDB: read excitonic kernel in transition space
        - YamboNLDB: read nonlinear response calculated with yambo_nl

    bse
        - YamboExcitonWaveFunctionXSF: read the excitonic
        - YamboExcitonWeight: read the excitonic weights from the ypp output file
        - YamboBSEAbsorptionSpectra: generate a .json file with the bse absorption calculation (including information about the excitons)

    em1s
        - YamboEm1sRotate: rotate em1s from IBZ to BZ
        - YamboEm1sExpand: expand em1s from unit cell to supercell [IN DEVELOPMENT]
    analyse:
        - YamboAnalyser: read .json files generated with yamboout and plot them together
"""
import numpy as np

#tools and units
from yambopy.tools.jsonencoder import *
from yambopy.tools.types import *
from yambopy.units import *
from yambopy.zeros import *

#lattice-related operations
from yambopy.lattice import *


#skw interpolator (adapted from abipy version)
from yambopy.tools.skw import *

#yambo databases
from yambopy.dbs.dipolesdb import *
from yambopy.dbs.qpdb import *
from yambopy.dbs.hfdb import *
from yambopy.dbs.em1sdb import *
from yambopy.dbs.greendb import *
from yambopy.dbs.latticedb import *
from yambopy.dbs.electronsdb import *
from yambopy.dbs.rtdb import *
from yambopy.dbs.rt_carriers_db import *
from yambopy.dbs.nldb import *
from yambopy.dbs.excitondb import *
from yambopy.dbs.wfdb import *
from yambopy.dbs.elphondb import *
from yambopy.dbs.bsekerneldb import *
from yambopy.dbs.excphondb import *
from yambopy.dbs.kqgridsdb import *

#kpoint mesh operations
from yambopy.kpoints import *

#input/output files
from yambopy.io.inputfile import *
from yambopy.io.outputfile import *
from yambopy.io.jsonfile import *
from yambopy.io.iofile import *
from yambopy.io.xsffile import *

#bse/excitons files
from yambopy.bse.excitonwf import *
from yambopy.bse.excitonweight import *
from yambopy.bse.bse_absorption import *
from yambopy.bse.bse_dispersion import *
from yambopy.bse.excitonradiativelifetimes import *
from yambopy.bse.excitondipoles import *

#em1s/static screening operations files
from yambopy.em1s.em1s_rotate import *

#ndb.QP operations
from yambopy.quasiparticles.QP_rotate import *

#LetzElPhC interface
from yambopy.letzelphc_interface.lelphcdb import *
from yambopy.letzelphc_interface.lelph2y import *

#analyse stuff
from yambopy.analyse import *

#workflow files
from yambopy.common.save_generation import *
from yambopy.common.workflow import *
from yambopy.common.calculation_manager import *
from yambopy.common.transform_matrix_element import *

#realtime files
from yambopy.rt.rt_movie import *
from yambopy.rt.rt_timestep_optimize import *

#non-linear files
from yambopy.nl.linear_optics import *
from yambopy.nl.fft_interp import *
from yambopy.nl.external_efield import *
from yambopy.nl.damp_it import *
# from yambopy.nl.harmonic_analysis import *
# from yambopy.nl.sum_frequencies import *
from yambopy.nl.hhg_tools import *
from yambopy.nl.nl_analysis import *
from yambopy.nl.sin_analysis import *
from yambopy.nl.freqmix_analysis import *
from yambopy.nl.pulse_analysis import *
#doublegrid files
from yambopy.double_grid.dg_convergence import *

#gkkp files
from yambopy.gkkp.compute_gkkp import *
from yambopy.gkkp.refine_gkkp import *
from yambopy.gkkp.descreen_gkkp import *

#data
from yambopy.data import *

#exciton-phonon 
from yambopy.exciton_phonon.excph_input_data import *
from yambopy.exciton_phonon.excph_matrix_elements import *
from yambopy.exciton_phonon.excph_luminescence import *



import numpy as np
from scipy.fft import ifftn,fftn,fft2

class FourierTransformer:
    """
    Class to handle FFT transformation. 
    """    
    def __init__(self, dlat,rlat,qpoints,fft_grid_size): #fft_grid_size can 3D,2D array
        '''
        yambopy FFT class. Perform the np.fft transform of 2D and 3D grids
        fft_grid_size can be a tuple, array or numpy array.
        '''
        
        #mode can be either real to direct or direct2real 'd2r'
        self.dlat = dlat # direct lattice
        self.rlat = rlat #reciprocal lattice
        self.qpoints = np.array(qpoints)
        self.fft_grid_size = fft_grid_size
        # Compute the number of G-vectors based on the fft_grid_size
        self.num_g_vectors = np.prod(fft_grid_size)

        # Check the dimensionality of qpoints and adjust for 2D case if necessary
        if self.qpoints.shape[1] == 2:
            self.dim = 2
        elif self.qpoints.shape[1] == 3:
            self.dim = 3
        # elif self.qpoints.shape[1] == 1:
        #     self.dim = 1            
        else:
            raise ValueError("Invalid dimensionality for qpoints. Must be 2D or 3D.")        

        if self.dim != len(fft_grid_size):
            raise ValueError("Mismatch between dimensionality of qpoints and fft_grid_size.")
        #number of G-vecotrs based on the qpoints

    def get_grid(self):
        # Create the grid for real space coordinates
        x,y,z = np.linspace(0,self.fft_grid_size[0],1),np.linspace(0,self.fft_grid_size[1],1),np.linspace(0,self.fft_grid_size[2],1)
        if self.dim == 2:
            X, Y = np.meshgrid(x, y)
            return X, Y
            #space_grid = np.column_stack((x.ravel(), y.ravel()))
        elif self.dim == 3:
            X, Y, Z = np.meshgrid(x, y, z)
            #space_grid = np.column_stack((x.ravel(), y.ravel(), z.ravel()))
            return X, Y, Z
    def real_to_reciprocal(self, data_real):
        # Perform the Fourier transform to get the function in reciprocal space
        data_reciprocal = fftn(data_real, s = self.fft_grid_size)
        return data_reciprocal


import numpy as np
import argparse

"""
Script to calculate the expected BSE kernel size, in GB, to be loaded by each MPI task

Size_kernel_GB = (nk*nv*nc*ncpl)**2*np*4*2)/(1024.)**3.

Usage:

    :: -nk,--nkpts_bz      = Number of kpoints in the full BZ
    :: -nv,--nvalence,     = Number of valence bands included in the kernel
    :: -nc,--nconduction   = Number of conduction bands included in the kernel
    :: -ndp,--ndoubleprec  = Double precision [Default: False]
    :: -ncpl,--ncoupling   = BSE kernel in coupling mode (i.e., no TDA) [Default: False ]

"""

def get_BSE_kernel_size(nk,nv,nc,np,ncpl):

    bytes_to_GB = (1024.)**3.
    float_bytes = 4.
    ncmplx = 2.

    BSE_kernel_dimension = nk*nv*nc
    BSE_kernel_size = (BSE_kernel_dimension*ncpl)**2.*np*float_bytes*ncmplx/bytes_to_GB

    print('============= BSE kernel size ============')
    print(f' Kernel dimension (resonant part): {BSE_kernel_dimension}')
    print( ' Expected full kernel size [GB]  : %.3f'%BSE_kernel_size)
    print('==========================================')

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Calculate expected size of BSE kernel to be loaded by each MPI task (in GB)')
    parser.add_argument('-nk',   '--nkpts_bz', type=int, help='Number of kpoints in the full BZ',required=True)
    parser.add_argument('-nv',   '--nvalence', type=int, help='Number of valence bands included in the kernel',required=True)
    parser.add_argument('-nc','--nconduction', type=int, help='Number of conduction bands included in the kernel',required=True)
    parser.add_argument('-ndp', '--ndoubleprec', type=bool, help='double precision',action='store_false')
    parser.add_argument('-ncpl','--ncoupling', type=bool, help='BSE kernel in coupling mode (i.e., no TDA)',action='store_false')

    args = parser.parse_args()

    nk   = args.nkpts_bz
    nv   = args.nvalence
    nc   = args.nconduction
    np   = 1 if args.ndoubleprec else 2
    ncpl = 1 if args.ncoupling else 2

    get_BSE_kernel_size(nk,nv,nc,np,ncpl)



# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
import os
import re
from collections import defaultdict

import numpy as np
from netCDF4 import Dataset

from .yambofile import YamboFile
from yambopy import YamboIn
from yambopy.tools.jsonencoder import JsonDumper
from yambopy.dbs.latticedb import YamboLatticeDB
from yambopy.units import ha2ev

class YamboOut():
    """ 
    Class to read yambo output files and pack them in a .json file

    **Arguments:**

    ``folder``:      The relative path of the folder where yambo dumped its input files

    ``save_folder``: The path were the SAVE folder is localized 
    """

    _tags = ['refl', 'eel', 'eps', 'qp', 'sf',
             'carriers', 'polarization', 'external']
    _netcdf = ['ndb.QP', 'ndb.HF_and_locXC']
    _tagsexp = r'#\n#\s+((?:(?:[`0-9a-zA-Z\-\/\|\\(\)\_[\]]+)\s+)+)#\n\s'

    def __init__(self, folder, save_folder='.'):
        """
        For now we initialize the class in the same way as before (from_files)
        In the future we should be able to initialize this class from
        a file in the disk or from a dictionary
        """
        self.files = defaultdict(dict)
        self.from_folder(folder,save_folder)

    def from_folder(self,folder,save_folder):
        self.folder = folder

        # check if the save folder is in save_folder if not try folder
        if not os.path.isdir(save_folder + '/SAVE'):
            self.save_folder = folder
        else:
            self.save_folder = save_folder

        # get all the files in the output dir
        if os.path.isdir(folder):
            outdir = os.listdir(folder)
        else:
            raise ValueError("Invalid folder: %s" % folder)

        # get the log dir
        logdir_path = os.path.join(folder, "LOG")
        if os.path.isdir(logdir_path):
            logdir = os.listdir(logdir_path)
        else:
            logdir = outdir

        # get output filenames
        self.netcdf = ["%s" % f for f in outdir if f in self._netcdf]
        self.output = ["%s" % f for f in outdir if f.startswith('o-') and YamboFile.has_tag(f,self._tags)]
        self.run = ["%s" % f for f in outdir if f.startswith('r-')]
        self.logs = ["%s" % f for f in logdir if f.startswith('l-')]

        # get data from output file
        self.get_runtime()
        self.get_outputfile()
        self.get_netcdffile()
        self.get_inputfile()
        self.get_cell()

    @staticmethod
    def has_output(folder):
        """Check if the folder has output files"""
        return [filename for filename in os.listdir(folder) if YamboFile.is_output(filename)] 

    def get_cell(self):
        """ 
        Get information about the unit cell (lattice vectors, atom types, positions,
        kpoints and symmetry operations) from the SAVE folder.
        """
        try:
            path = os.path.join(self.save_folder,'SAVE/ns.db1')
            self.lattice = YamboLatticeDB.from_db_file(path)
        except: #AiiDA
            path = self.save_folder+'/ns.db1'
            self.lattice = YamboLatticeDB.from_db_file(path)

        

    def get_outputfile(self):
        """ 
        Get the data from the o-* files
        """
        # for all the o-* files
        for filename in self.output:
            
            #yambofile read
            # TODO: use Yambofile class to read the o-* files
            yf = YamboFile(filename,self.folder)

            #classic read
            with open(os.path.join(self.folder, filename),'r') as f:
                string = f.read()

                # get tags
                find_tags = re.findall(self._tagsexp, string)
                tags = [tag.strip() for tag in find_tags[0].split()]
                f.seek(0)

                # get data
                data = np.loadtxt(f, unpack=True)

                #store data
                self.files[filename].update(dict(zip(tags,data)))
                self.files[filename]["type"] = yf.type

    def get_netcdffile(self):
        """
        Get the netcdf files
            The supported netcdf files so far are:
                ndb.QP
                ndb.HF_and_locXC
        """
        for filename in self.netcdf:
            yf = YamboFile(filename, self.folder)
            #convert units
            if yf.type == 'netcdf_gw':
                yf.data['E-Eo'] *= ha2ev
                yf.data['Eo'] *= ha2ev
                yf.data['E'] *= ha2ev
            #save data
            self.files[filename] = yf.data
            self.files[filename]["type"] = yf.type

    def get_inputfile(self):
        """
        Get the input file from the o-* file
        """

        for filename in self.output:
            inputfile = []
            # read this inputfile
            with open(os.path.join(self.folder, filename),'r') as f:
                for line in f:
                    if 'Input file :' in line:
                        for line in f:
                            # Note this: to read the input file we just ignore the 
                            # first 4 characters of the section after the tag 'Input file:'
                            inputfile.append(line[4:])

            # use YamboIn to read the input file to a list
            yi = YamboIn()
            self.files[filename]["input"] = yi.read_string(''.join(inputfile))

    def get_runtime(self):
        """
        Get the runtime from the r-* file
        """

        for filename in self.run:
            timing = {}

            with open(os.path.join(self.folder,filename),'r') as f:

                category = "UNKNOWN"
                for line in f:
                    if 'Timing' in line:
                        timing[category] = line.split()[-1].split('/')
                    if re.search('(\[[0-9.]+\].[A-Z])', line):
                        category = line.strip()
                
            self.files[filename]["runtime"] = timing
            self.files[filename]["type"] = "report"

    def get_data(self, tags):
        """
        Search for a tag in the output files and obtain the data
        """
        data = {}
        for key in self.files.keys():
            if all(tag in key for tag in tags):
                data[key] = self.files[key]
        return data

    def print_runtime(self):
        """
        Print the runtime in a string
        """
        timing = self.get_runtime()
        for t in list(timing.items()):
            print(t[0], '\n', t[1], '\n')

    def pack(self, filename=None):
        """
        Pack up all the data in the structure in a json file
        """
        # if no filename is specified we use the same name as the folder
        if not filename:
            filename = '%s.json' % self.folder

        # create json dictionary
        jsondata = {"files": self.files,
                    "lattice": self.lattice.as_dict()}
        JsonDumper(jsondata, filename)

    def __str__(self):
        lines = []
        lines.append("logs:")
        lines += self.logs
        lines.append("\nrun:")
        lines += self.run
        lines.append("\noutput (text):")
        lines += self.output
        lines.append("\noutput (netcdf):")
        lines += self.netcdf
        lines += [str(self.lattice)]
        return "\n".join(lines)



# Copyright (C) 2016 Henrique Pereira Coutada Miranda, Alejandro Molina-Sanchez
# All rights reserved.
#
# This file is part of yambopy
#
#
import os
import subprocess
from .scheduler import Scheduler

class Slurm(Scheduler):
    """
    Class to submit jobs through the Slurm scheduler

    ``_vardict`` states the default assignement of the nodes and cores variables
    from the schduler class to the variables needed in this class
    """
    _vardict = {"cores":"core",
                "nodes":"nodes",
                "cpus_per_task":"cpus_per_task"}
                          
                
    def initialize(self):
        self.get_vardict()

    def get_script(self):
        """
        get a .sh file to be submitted using sbatch
        sbatch <filename>.sh
        """
        lines = []; app = lines.append
        app('#!/bin/bash -l\n')
        
        partition = self.get_arg("partition",None)
        if self.name: app("#SBATCH -J %s"%self.name)
        if partition: app('#SBATCH --partition %s'%partition)

        qos = self.get_arg("qos",None)
        if qos: app('#SBATCH --qos=%s'%qos)

        dependency = self.get_arg("dependency",None)
        if dependency: app("#SBATCH --dependency=%s"%dependency)

        if self.nodes: app("#SBATCH -N %d" % self.nodes)
        if self.cores: app("#SBATCH --ntasks-per-node=%d" % self.cores)
        if self.cpus_per_task: app("#SBATCH --cpus-per-task=%d" % self.cpus_per_task )

        mem_per_cpu = self.get_arg("mem_per_cpu",None) 
        if mem_per_cpu: app("#SBATCH --mem-per-cpu=%s" % mem_per_cpu)

        app("#SBATCH --time=0-%s" % self.walltime)

        app(self.get_commands())
        return "\n".join(lines)

    def __str__(self):
        """
        create the string for this job
        """
        return self.get_script()

    def run(self,filename='run.sh',dry=False,command="sbatch",verbose=0):
        """
        Create the submission script and submit the job
        
        Arguments:
            dry - only print the commands to be run on the screen
        """
        if dry: 
            print(self)
            return 

        #create the submission script
        self.write(filename)        
        workdir  = os.path.dirname(filename)
        basename = os.path.basename(filename) 

        p = subprocess.Popen([command,basename],stdout=subprocess.PIPE,stderr=subprocess.PIPE,cwd=workdir)
        self.stdout,self.stderr = p.communicate()
        #Slurm-specific instruction to store jobid
        self.jobid = self.stdout.decode().split(' ')[-1].strip()
        
        #check if there is stderr
        if self.stderr: raise Exception(self.stderr)
        
        #check if there is stdout
        if verbose: print(self.stdout)
        
    def check_job_status(self,workdir):
        """
        Return status of slurm job (empty if job is not present)
        """
        p = subprocess.Popen(['squeue','-j %s'%self.jobid],stdout=subprocess.PIPE,stderr=subprocess.PIPE,cwd=workdir)
        stdout,stderr = p.communicate()
        if stdout:
            if stdout.decode()[-9:-1]=='(REASON)': job_status = 'NULL' 
            else: job_status = stdout.decode().split('\n')[1].split()[4]
        else: 
            job_status = 'NULL'
        return job_status



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: HPC
#
# This file is part of the yambopy project
#
import numpy as np
from scipy.interpolate import griddata
import matplotlib.pyplot as plt
from yambopy.lattice import red_car
from yambopy.kpoints import get_path
from yambopy.dbs.latticedb import YamboLatticeDB
from itertools import product
from netCDF4 import Dataset

class YamboExcitonWeight(object):
    """
    Class to read the excitonic weight writen by ypp
    """

    def __init__(self,filename,save='SAVE',path='.'):
        #read save database

        # ! I have commented line 20 because is was broken. Pls check

        ysave = YamboLatticeDB.from_db_file(folder=save,filename='ns.db1')
        self.sym_car  = ysave.sym_car
        self.kpts_car = ysave.car_kpoints #kpts_car
        self.rlat     = ysave.rlat
        self.lat = ysave.lat

        #super(YamboLatticeDB,self).from_db_file(folder=save,filename='ns.db1')

        #read excitons file
        self.excitons = np.loadtxt(filename)

        self.weights = None

    def write_irr(self,filename="irr.dat"):
        """ write the list of kpoints from the irreducible brillouin zone
        """
        f = open("irr.dat",'w')
        for k in self.kpts_car:
            f.write(("%12.8lf"*3)%(k[1],k[0],k[2])+"\n")
        f.close()

    def write_full(self,filename="full.dat"):
        """ write the list of kpoints in the full brillouin zone
        """
        #generate all the possible points
        kpoints = self.kpts_car
        kpts = []
        for k in self.kpts_car:
            for sym in self.sym_car:
                kpts.append(np.dot(sym,k))

        f = open("full.dat",'w')
        for q in kpts:
            f.write(("%12.8lf "*3)%tuple(q)+"\n")
        f.close()

    def get_data(self):
        qpts, weights, transitions = self.calc_kpts_weights()
        return { "qpts": qpts,
                 "weights": weights,
                 "lattice": self.lat,
                 "reciprocal_lattice": self.rlat,
                 "transitions": transitions
                 }

    def calc_kpts_weights(self,repx=list(range(-1,2)),repy=list(range(-1,2)),repz=list(range(-1,2))):
        """ Calculate the weights and kpoints of the excitons
        """
        self.weights     = dict()

        #first run set everything to zero
        for line in self.excitons:
            v,c,k,sym,w,e = line
            self.weights[(int(k),int(sym))] = 0

        #add weights
        for line in self.excitons:
            v,c,k,sym,w,e = line
            self.weights[(int(k),int(sym))] += w

        #rename symmetries and kpoints
        sym = self.sym_car
        kpoints = self.kpts_car

        qpts     = []
        kidx     = []
        weights  = []

        for r in product(repx,repy,repz):
          for k,s in list(self.weights.keys()):
            w   = self.weights[(k,s)]
            weights.append( w )
            qpt = np.dot(sym[s-1],kpoints[k-1])+red_car([r],self.rlat)[0]
            qpts.append( qpt )
            kidx.append( k )

        return np.array(qpts), np.array(weights)

    def calc_kpts_transitions(self,repx=list(range(-1,2)),repy=list(range(-1,2)),repz=list(range(-1,2)),debug=False):
        """ Calculate the transitions and kpoints of the excitons
        """
        self.weights     = dict()
        self.transitions = dict()
        self.transitions_v_to_c = dict()

        #first run set everything to zero
        for line in self.excitons:
            v,c,k,sym,w,e = line
            self.weights[(int(k),int(sym))] = 0
            self.transitions[(int(v),int(c),int(k),int(sym))] = 0
            self.transitions_v_to_c[(int(v),int(c))] = 0

        #add weights
        for line in self.excitons:
            v,c,k,sym,w,e = line
            self.weights[(int(k),int(sym))] += w

        #add transitions
        for line in self.excitons:
            v,c,k,sym,w,e = line
            self.transitions[(int(v),int(c),int(k),int(sym))] += w

        #add percentage of a given v => c transition
        norm = sum(self.excitons[:,4])
        for v,c,k,s in list(self.transitions.keys()):
          self.transitions_v_to_c[(int(v),int(c))] += self.transitions[(v,c,k,s)]
        if debug: print('transitions (valence > condution):')
        for v,c in self.transitions_v_to_c:
          self.transitions_v_to_c[(v,c)] = self.transitions_v_to_c[(v,c)]/norm
          if debug: print('%3d > %3d'%(v,c))

        #rename symmetries and kpoints
        sym = self.sym_car
        kpoints = self.kpts_car

        qpts     = []
        kidx     = []
        t_v_c    = []

        for r in product(repx,repy,repz):
          for k,s in list(self.weights.keys()):
            qpt = np.dot(sym[s-1],kpoints[k-1])+red_car([r],self.rlat)[0]
            qpts.append( qpt )
            kidx.append( k )
            #print (v_ref,c_ref,k,s)
            #aux.append(self.transitions[(v_ref,c_ref,k,s)])

        for v_ref,c_ref in list(self.transitions_v_to_c.keys()):
          aux = []
          for r in product(repx,repy,repz):
            for k,s in list(self.weights.keys()):
              aux.append(self.transitions[(v_ref,c_ref,k,s)])
          t_v_c.append(np.array(aux))

        return np.array(qpts), t_v_c, np.array(kidx)

    def plot_contour(self,resX=500,resY=500):
        """ plot a contour
            idea taken from http://stackoverflow.com/questions/18764814/make-contour-of-scatter
        """
        kpts, z = self.calc_kpts_weights()
        x,y = kpts[:,0],kpts[:,1]
        xi = np.linspace(min(x), max(x), resX)
        yi = np.linspace(min(y), max(y), resY)
        Z = griddata(x, y, z, xi, yi, interp='cubic')
        X, Y = np.meshgrid(xi, yi)

        plt.contourf(X, Y, Z, cmap='gist_heat_r')
        plt.show()

    def plot_weights(self,ax,size=20,marker='H',set_origin=0.0,lim=0.2,cmap='viridis',log_scale=False,set_maximum=1.0):
        """

        Plot the weights in a scatter plot of this exciton (1st version tuned by A. Molina-Sanchez)
        Options:
        cmap : colormap. Default viridis 
        log_scale : Logarithmic scale for the intensity (True or False)
        set_maximum : Only applied for linear scale. Apply a cut for a selected intensity (values between 0 and 1)
        Further development: Option for the colorbar
 
        """
        from numpy import sqrt
        import matplotlib.pyplot as plt
        import matplotlib.colors as colors

        """
        These options can be decided by the user. In this first version we just:
        remove axis
        """
        ax.set_aspect('equal')
        ax.set_xlim(-lim,lim)
        ax.set_ylim(-lim,lim)
        ax.axes.get_xaxis().set_visible(False)
        ax.axes.get_yaxis().set_visible(False)

        
        kpts, weights = self.calc_kpts_weights()

        if log_scale == True:
           norm = colors.LogNorm(vmin=weights.min(),vmax=weights.max())
        else:
           if abs(set_maximum)>1.:
              set_maximum = 1.
           norm = colors.Normalize(vmin=weights.min(),vmax=abs(set_maximum)*weights.max())

        cmap = plt.get_cmap(cmap)

        ax.scatter(kpts[:,0]-set_origin, kpts[:,1]-set_origin, s=size, marker=marker, color=cmap(norm(weights)))

    def __str__(self):
        s = ""
        s += "reciprocal lattice:\n"
        s += "\n".join([("%12.8lf "*3)%tuple(r) for r in self.rlat])+"\n"
        s += "lattice:\n"
        s += "\n".join([("%12.8lf "*3)%tuple(r) for r in self.lat])+"\n"
        s += "alat:\n"
        s += ("%12.8lf "*3)%tuple(self.alat)+"\n"
        return s

    def plot_transitions(self,size=30,lim=0.2):
        """
        Plot the weight of a given transition in a scatter plot of this exciton.
        My idea is to associate for each transition a color and to plot in a different plot
        """

        from numpy import sqrt
        cmap = plt.get_cmap("gist_heat_r")

        fig = plt.figure(figsize=(10,10))
        kpts, t_v_c, _ = self.calc_kpts_transitions()
        for individual in t_v_c:
          plt.scatter(kpts[:,0], kpts[:,1], s=size, marker='H', color=[cmap(sqrt(c)) for c in individual])

        plt.xlim([-lim,lim])
        plt.ylim([-lim,lim])
        ax = plt.axes()
        ax.set_aspect('equal')
        plt.show()

    def plot_exciton_bs(self,ax,path,nbands='all',space='transition',color='#1f77b4'):
        """
        Plot the excitonic weights of a given transition in the band-structure
        """
        kpts, t_v_c, kidx = self.calc_kpts_transitions(repx=list(range(1)),repy=list(range(1)),repz=list(range(1)))
        t_v_c = np.array(t_v_c)

        #get_path is provided by savedb
        bands_kpoints, bands_indexes, path_car = get_path(kpts,self.rlat,None,path)

        #calculate distances
        bands_distances = [0]
        distance = 0
        for nk in range(1,len(bands_kpoints)):
            distance += np.linalg.norm(bands_kpoints[nk-1]-bands_kpoints[nk])
            bands_distances.append(distance)

        #get energies at these k-points
        eig = self.eigenvalues[kidx-1]
        eig = eig[bands_indexes]
        transition_weight  = t_v_c[:,bands_indexes]
        if nbands == 'all': nbands = self.nbands
        for tw,t in zip(transition_weight,list(self.transitions_v_to_c.keys())):
            v,c = t
            if space == 'transition':
                ax.plot(bands_distances,eig[:,c-1]-eig[:,v-1])
                ax.scatter(bands_distances,eig[:,c-1]-eig[:,v-1],s=tw*1e4)
            else:
                ax.plot(bands_distances,eig[:,c-1],c=color)
                ax.plot(bands_distances,eig[:,v-1],c=color)
                ax.scatter(bands_distances,eig[:,c-1],s=tw*1e4,c=color)
                ax.scatter(bands_distances,eig[:,v-1],s=tw*1e4,c=color)

    def __str__(self):
        s = ""
        s += "reciprocal lattice:\n"
        s += "\n".join([("%12.8lf "*3)%tuple(r) for r in self.rlat])+"\n"
        s += "lattice:\n"
        s += "\n".join([("%12.8lf "*3)%tuple(r) for r in self.lat])+"\n"
        s += "alat:\n"
        s += ("%12.8lf "*3)%tuple(self.alat)+"\n"
        return s



import os
from yambopy import *
from schedulerpy import *
import argparse

"""
Script to produce a yambo SAVE folder.
Usable to quickly initialize a standard yambo calculation.

Inputs:
 1. --nscf_dir='path/to/nscf/save/folder'
 2. --yambo='path/to/yambo/executables' [OPTIONAL]

NB:
 The SAVE folder is created in the directory where the script is launched!

"""

def generate_save(database,qe_save,y_dir,scheduler,noinit=False):
    """
    Generate SAVE folder from QE nscf calculation
    """
    #check if the nscf cycle is present
    if os.path.isdir(qe_save):
        print('nscf calculation found!')
    else:
        print('nscf calculation not found!')
        exit()

    if y_dir!="": y_dir=y_dir+"/"
    p2y = y_dir+"p2y"
    yambo = y_dir+"yambo"

    #check if the SAVE folder is present
    if os.path.isdir('%s/SAVE'%database):
        print('SAVE database found!')
    if not os.path.isdir('%s/SAVE'%database):
        print('preparing yambo database')

        p2y_run = scheduler()
        p2y_run.add_command('mkdir -p %s'%database)
        p2y_run.add_command('cd %s; %s > p2y.log ; cd -'%(qe_save,p2y))
        if not noinit: p2y_run.add_command('cd %s; %s > yambo.log ; cd -'%(qe_save,yambo))
        p2y_run.add_command('mv %s/SAVE %s'%(qe_save,database))
        p2y_run.run()
        
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Generate SAVE folder')
    parser.add_argument('-nscf','--nscf_dir', type=str,help='<Required> Path to nscf save folder', required=True)
    parser.add_argument('-y','--yambo_dir', type=str, default="", help='<Optional> Path to yambo executables')
    args = parser.parse_args()

    nscf_dir = args.nscf_dir
    yambo_dir = args.yambo_dir

    database = './'
    scheduler = Scheduler.factory

    generate_save(database,nscf_dir,yambo_dir,scheduler)




#
# Authors: MN
###
# This file contains a genenal functions to compute
# < S | O | S'>, where O is an operator.
# for Ex: is O is dV_scf, then these ,matrix elements are 
# ex-ph matrix elements. if O is S_z (spin operator), we get
# spin matrix elements of excitons
###
import numpy as np
from yambopy.kpoints import build_ktree, find_kpt
from yambopy.tools.function_profiler import func_profile

@func_profile
def exciton_X_matelem(exe_kvec, O_qvec, Akq, Ak, Omn, kpts, contribution='b', diagonal_only=False, ktree=None):
    """
    Compute the exciton matrix elements in the Tamm-Dancoff approximation.

    This function calculates the matrix elements <S (final), k+q | O(q) | S' (initial), k>,
    discarding the third term (disconnected diagram). The calculation is performed in the
    Tamm-Dancoff approximation.

    Parameters
    ----------
    exe_kvec : array_like
        Exciton k-vector in crystal coordinates (k).
    O_qvec : array_like
        Momentum transfer vector q in crystal coordinates (q).
    Akq : array_like
        Wavefunction coefficients for k+q (bra wfc) with shape (n_exe_out, 1, ns, nk, nc, nv).
    Ak : array_like
        Wavefunction coefficients for k (ket wfc) with shape (n_exe_in, 1, ns, nk, nc, nv). [NB: in luminescence, this is the Q=0 intermediates state]
    Omn : array_like
        Matrix elements of the operator O in the basis of electronic states with shape (nlambda, nk, nspin, m_bnd, n_bnd).
        ie Omn = < k+q, m, s | O(q) | n, k, s>, where m_bnd and n_bnd are final and initial bands, respectively.
        s is spin index
    kpts : array_like
        K-points used to construct the BSE with shape (nk, 3) in crystal coordinates.
    contribution : str, optional
        Specifies the contribution to include in the calculation:
        - 'e' : Only electronic contribution.
        - 'h' : Only hole contribution.
        - 'b' : Both electron and hole contributions (default).
    diagonal_only : bool, optional
        If True, only the diagonal terms are computed. Default is False.

    ktree : KDtree, optional
        If None, will build internally, else use the user provided
    Returns
    -------
    ex_O_mat : ndarray
        The computed exciton matrix elements with shape (nlambda, n_exe_out) if diagonal_only is True,
        or (nlambda, n_exe_out (final), n_exe_in (initial)) if diagonal_only is False.
    """
    # Number of arbitrary parameters (lambda) in the Omn matrix (e.g., phonon modes)
    nlambda = Omn.shape[0]
    #
    assert Akq.shape[1] == 1, "Works only with TDA."
    # Shape of the wavefunction coefficients
    n_exe_out, bse_calc, ns, nk, nc, nv = Akq.shape
    n_exe_in                            = Ak.shape[0]
    if n_exe_in < n_exe_out:
        print(f"[WARNING] less intermediate states {n_exe_in} than final states {n_exe_out}")
    #
    # Ensure that the shapes of Akq and Ak match
    assert Akq.shape[1:] == Ak.shape[1:], "Wavefunction coefficient mismatch"
    #
    # Ensure that the contribution parameter is valid
    assert contribution in ['b', 'e', 'h'], "Allowed values for contribution are 'b', 'e', 'h'"
    #
    # Build a k-point tree for efficient k-point searching
    if ktree is None : ktree = build_ktree(kpts)
    #
    # Find the indices of k-Q-q and k-q in the k-point tree
    idx_k_minus_Q_minus_q = find_kpt(ktree, kpts - O_qvec[None, :] - exe_kvec[None, :])  # k-Q-q
    idx_k_minus_q = find_kpt(ktree, kpts - O_qvec[None, :])  # k-q
    #
    # Extract the occupied and unoccupied parts of the Omn matrix
    Occ = Omn[:, idx_k_minus_q, :, nv:, nv:].transpose(0,2,1,3,4)  # conduction (electron) part
    Ovv = Omn[:, idx_k_minus_Q_minus_q, :, :nv, :nv].transpose(0,2,1,3,4)  # valence (hole) part
    #
    # Ensure the arrays are C-contiguous to reduce cache misses
    Ak_electron = np.ascontiguousarray(Ak[:,0][:,:,idx_k_minus_q, ...])
    Akq_conj = Akq[:,0].reshape(n_exe_out, -1).conj()
    #
    # Initialize the output matrix
    if diagonal_only:
        ex_O_mat = np.zeros((nlambda, n_exe_out), dtype=Ak.dtype)  # (nlambda, final, initial)
    else:
        ex_O_mat = np.zeros((nlambda, n_exe_out, n_exe_in), dtype=Ak.dtype)  # (nlambda, final, initial)
    #
    # Loop over the arbitrary parameters (lambda)
    for il in range(nlambda):
        # Compute the electron contribution
        if contribution == 'e' or contribution == 'b':
            tmp_wfc = Occ[il][None, ...] @ Ak_electron
        #
        # Compute the hole contribution and subtract from the electron contribution
        if contribution == 'h' or contribution == 'b':
            tmp_h = -Ak[:,0] @ Ovv[il][None, ...]
            if contribution == 'b':
                tmp_wfc += tmp_h
            else:
                tmp_wfc = tmp_h
        #
        # Reshape the temporary wavefunction coefficients
        tmp_wfc = tmp_wfc.reshape(n_exe_in, -1)
        #
        # Compute the final matrix elements
        if diagonal_only:
            ex_O_mat[il] = np.sum(Akq_conj * tmp_wfc, axis=-1)
        else:
            np.matmul(Akq_conj, tmp_wfc.T, out=ex_O_mat[il])
    #
    # Return the computed exciton matrix elements
    return ex_O_mat

